"""
===================
Generic Model class
===================

**Base class of all implementations.**

Inherit to *implement*, i.e. to define support for a specific model type.

Definitions
-----------
"""
from __future__ import annotations

import dataclasses
import json
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Optional, Union

import pharmpy
from pharmpy.internals.df import hash_df_runtime
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.immutable import Immutable, cache_method, frozenmapping
from pharmpy.model.external import detect_model

from .datainfo import ColumnInfo, DataInfo
from .estimation import EstimationSteps
from .parameters import Parameters
from .random_variables import RandomVariables
from .statements import CompartmentalSystem, Statements

if TYPE_CHECKING:
    import pandas as pd
    import sympy
else:
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy


class ModelError(Exception):
    """Exception for errors in model object"""

    pass


class ModelSyntaxError(ModelError):
    """Exception for Syntax errors in model code"""

    def __init__(self, msg='model syntax error'):
        super().__init__(msg)


class ModelfitResultsError(ModelError):
    """Exception for issues with ModelfitResults"""

    pass


@dataclass(frozen=True)
class ModelInternals:
    def __init__(self):
        pass

    def replace(self, **kwargs):
        return dataclasses.replace(self, **kwargs)


class Model(Immutable):
    """The Pharmpy model class"""

    def __init__(
        self,
        name: str = '',
        parameters: Parameters = Parameters(),
        random_variables: RandomVariables = RandomVariables.create(()),
        statements: Statements = Statements(),
        dataset: Optional[pd.DataFrame] = None,
        datainfo: DataInfo = DataInfo(),
        dependent_variables: frozenmapping[sympy.Symbol, int] = frozenmapping(
            {sympy.Symbol('y'): 1}
        ),
        observation_transformation: Optional[frozenmapping[sympy.Symbol, sympy.Expr]] = None,
        estimation_steps: EstimationSteps = EstimationSteps(),
        modelfit_results=None,
        parent_model: Optional[str] = None,
        initial_individual_estimates: Optional[pd.DataFrame] = None,
        filename_extension: str = '',
        value_type: str = 'PREDICTION',
        description: str = '',
        internals: Optional[ModelInternals] = None,
    ):
        self._name = name
        self._datainfo = datainfo
        self._dataset = dataset
        self._random_variables = random_variables
        self._parameters = parameters
        self._statements = statements
        self._dependent_variables = dependent_variables
        if observation_transformation is None:
            observation_transformation = frozenmapping(
                {dv: dv for dv in dependent_variables.keys()}
            )
        self._observation_transformation = observation_transformation
        self._estimation_steps = estimation_steps
        self._modelfit_results = modelfit_results
        self._parent_model = parent_model
        self._initial_individual_estimates = initial_individual_estimates
        self._filename_extension = filename_extension
        self._value_type = value_type
        self._description = description
        self._internals = internals

    @classmethod
    def create(
        cls,
        name: str,
        parameters: Optional[Parameters] = None,
        random_variables: Optional[RandomVariables] = None,
        statements: Optional[Statements] = None,
        dataset: Optional[pd.DataFrame] = None,
        datainfo: DataInfo = DataInfo(),
        dependent_variables: Optional[Mapping[Union[str, sympy.Symbol], int]] = None,
        observation_transformation: Optional[Mapping[Union[str, sympy.Symbol], sympy.Expr]] = None,
        estimation_steps: Optional[EstimationSteps] = None,
        modelfit_results=None,
        parent_model: Optional[str] = None,
        initial_individual_estimates: Optional[pd.DataFrame] = None,
        filename_extension: str = '',
        value_type: str = 'PREDICTION',
        description: str = '',
        internals: Optional[ModelInternals] = None,
    ):
        Model._canonicalize_name(name)
        dvs = Model._canonicalize_dependent_variables(dependent_variables)
        obs_transformation = Model._canonicalize_observation_transformation(
            observation_transformation, dvs
        )
        parameters = Model._canonicalize_parameters(parameters)
        random_variables = Model._canonicalize_random_variables(random_variables)
        parameters = Model._canonicalize_parameter_estimates(parameters, random_variables)
        estimation_steps = Model._canonicalize_estimation_steps(estimation_steps)
        value_type = Model._canonicalize_value_type(value_type)
        if not isinstance(datainfo, DataInfo):
            raise TypeError("model.datainfo must be of DataInfo type")

        if dataset is not None:
            datainfo = update_datainfo(datainfo, dataset)

        statements = Model._canonicalize_statements(
            statements, parameters, random_variables, datainfo
        )
        return cls(
            name=name,
            dependent_variables=dvs,
            observation_transformation=obs_transformation,
            parameters=parameters,
            random_variables=random_variables,
            estimation_steps=estimation_steps,
            statements=statements,
            modelfit_results=modelfit_results,
            description=description,
            parent_model=parent_model,
            filename_extension=filename_extension,
            value_type=value_type,
            internals=internals,
            initial_individual_estimates=initial_individual_estimates,
            dataset=dataset,
            datainfo=datainfo,
        )

    @staticmethod
    def _canonicalize_value_type(value: str) -> str:
        allowed_strings = ('PREDICTION', 'LIKELIHOOD', '-2LL')
        if isinstance(value, str):
            if value.upper() not in allowed_strings:
                raise ValueError(
                    f"Cannot set value_type to {value}. Must be one of {allowed_strings} "
                    f"or a symbol"
                )
            value = value.upper()
        elif not isinstance(value, sympy.Symbol):
            raise ValueError("Can only set value_type to one of {allowed_strings} or a symbol")
        return value

    @staticmethod
    def _canonicalize_parameter_estimates(params, rvs) -> Parameters:
        inits = params.inits
        if not rvs.validate_parameters(inits):
            nearest = rvs.nearest_valid_parameters(inits)
            before, after = compare_before_after_params(inits, nearest)
            warnings.warn(
                f"Adjusting initial estimates to create positive semidefinite "
                f"omega/sigma matrices.\nBefore adjusting:  {before}.\n"
                f"After adjusting: {after}"
            )
            params = params.set_initial_estimates(nearest)
        return params

    @staticmethod
    def _canonicalize_random_variables(rvs: Optional[RandomVariables]) -> RandomVariables:
        if not isinstance(rvs, RandomVariables):
            raise TypeError("model.random_variables must be of RandomVariables type")
        if rvs is None:
            return RandomVariables.create()
        else:
            return rvs

    @staticmethod
    def _canonicalize_statements(
        statements: Optional[Statements],
        params: Parameters,
        rvs: RandomVariables,
        datainfo: DataInfo,
    ) -> Statements:
        if statements is None:
            return Statements()
        if not isinstance(statements, Statements):
            raise TypeError("model.statements must be of Statements type")
        colnames = {sympy.Symbol(colname) for colname in datainfo.names}
        symbs_all = rvs.free_symbols.union(params.symbols).union(colnames)
        if statements.ode_system is not None:
            symbs_all = symbs_all.union({statements.ode_system.t})
        sset_prev = []
        for i, statement in enumerate(statements):
            if isinstance(statement, CompartmentalSystem):
                continue

            symbs = statement.expression.free_symbols
            if not symbs.issubset(symbs_all):
                # E.g. after solve_ode_system
                if isinstance(statement.symbol, sympy.Function):
                    symbs_all.add(sympy.Symbol(statement.symbol.func.__name__))
                    continue

                for symb in symbs:
                    if symb in symbs_all:
                        continue
                    if str(symb) == 'NaN':
                        continue
                    if statements.find_assignment(symb) is None:
                        if symb == sympy.Symbol('t'):
                            # FIXME: Hack because t is not defined on model
                            continue
                        if statements.ode_system and symb in statements.ode_system.amounts:
                            continue
                        raise ValueError(f'Symbol {symb} is not defined')
                    if Statements(sset_prev).find_assignment_index(symb) is None:
                        raise ValueError(f'Symbol {symb} defined after being used')

            sset_prev += statement
        return statements

    @staticmethod
    def _canonicalize_name(name: str) -> None:
        if not isinstance(name, str):
            raise TypeError("Name of a model has to be of string type")

    @staticmethod
    def _canonicalize_dependent_variables(
        dvs: Optional[Mapping[Union[str, sympy.Symbol], int]]
    ) -> frozenmapping[sympy.Symbol, int]:
        if dvs is None:
            dvs = {sympy.Symbol('y'): 1}
        for key, value in dvs.items():
            if isinstance(key, str):
                key = sympy.Symbol(key)
            if not isinstance(key, sympy.Symbol):
                raise TypeError("Dependent variable keys must be of string or sympy.Symbol type")
            if not isinstance(value, int):
                raise TypeError("Dependent variable values must be of int type")
        return frozenmapping(dvs)

    @staticmethod
    def _canonicalize_observation_transformation(
        obs: Optional[Mapping[Union[str, sympy.Symbol], Union[str, sympy.Expr]]],
        dvs: frozenmapping[sympy.Symbol, int],
    ) -> frozenmapping[sympy.Symbol, sympy.Expr]:
        if obs is None:
            obs = {dv: dv for dv in dvs.keys()}
        for key, value in obs.items():
            if isinstance(key, str):
                key = sympy.Symbol(key)
            if isinstance(value, str):
                value = parse_expr(value)
            if not isinstance(key, sympy.Symbol):
                raise TypeError(
                    "Observation transformation keys must be of string or sympy.Symbol type"
                )
            if not isinstance(value, sympy.Expr):
                raise TypeError("Observation transformation keys values must be of sympy.Expr type")
        return frozenmapping(obs)

    @staticmethod
    def _canonicalize_parameters(params: Optional[Parameters]) -> Parameters:
        if params is None:
            return Parameters()
        else:
            if not isinstance(params, Parameters):
                raise TypeError("parameters must be of Parameters type")
            return params

    @staticmethod
    def _canonicalize_estimation_steps(steps: Optional[EstimationSteps]) -> EstimationSteps:
        if steps is None:
            return EstimationSteps()
        else:
            if not isinstance(steps, EstimationSteps):
                raise TypeError("model.estimation_steps must be of EstimationSteps type")
            return steps

    def replace(self, **kwargs) -> Model:
        name = kwargs.get('name', self.name)
        Model._canonicalize_name(name)

        if 'dependent_variables' in kwargs:
            dependent_variables = Model._canonicalize_dependent_variables(
                kwargs['dependent_variables']
            )
        else:
            dependent_variables = self.dependent_variables

        if 'observation_transformation' in kwargs:
            observation_transformation = Model._canonicalize_observation_transformation(
                kwargs['observation_transformation'], dependent_variables
            )
        else:
            observation_transformation = self.observation_transformation

        if 'parameters' in kwargs:
            parameters = Model._canonicalize_parameters(kwargs['parameters'])
        else:
            parameters = self.parameters

        if 'random_variables' in kwargs:
            random_variables = Model._canonicalize_random_variables(kwargs['random_variables'])
        else:
            random_variables = self.random_variables

        parameters = Model._canonicalize_parameter_estimates(parameters, random_variables)

        if 'dataset' in kwargs:
            dataset = kwargs['dataset']
            new_dataset = True
        else:
            dataset = self._dataset
            new_dataset = False

        if 'datainfo' in kwargs:
            datainfo = kwargs['datainfo']
            if not isinstance(datainfo, DataInfo):
                raise TypeError("model.datainfo must be of DataInfo type")
            new_datainfo = True
        else:
            datainfo = self._datainfo
            new_datainfo = False

        if new_dataset and dataset is not None:
            datainfo = update_datainfo(datainfo, dataset)
            if not new_datainfo:
                datainfo = datainfo.replace(path=None)

        # Has to be checked after datainfo is updated since it looks for symbols in datainfo as well
        if 'statements' in kwargs:
            statements = Model._canonicalize_statements(
                kwargs['statements'], parameters, random_variables, datainfo
            )
        else:
            statements = self.statements

        if 'estimation_steps' in kwargs:
            estimation_steps = Model._canonicalize_estimation_steps(kwargs['estimation_steps'])
        else:
            estimation_steps = self.estimation_steps

        modelfit_results = kwargs.get('modelfit_results', self.modelfit_results)
        parent_model = kwargs.get('parent_model', self.parent_model)
        initial_individual_estimates = kwargs.get(
            'initial_individual_estimates', self.initial_individual_estimates
        )
        filename_extension = kwargs.get('filename_extension', self.filename_extension)
        if not isinstance(filename_extension, str):
            raise TypeError("Filename extension has to be of string type")
        if 'value_type' in kwargs:
            value_type = Model._canonicalize_value_type(kwargs['value_type'])
        else:
            value_type = self.value_type
        description = kwargs.get('description', self.description)
        internals = kwargs.get('internals', self._internals)
        return self.__class__(
            name=name,
            dependent_variables=dependent_variables,
            parameters=parameters,
            random_variables=random_variables,
            statements=statements,
            dataset=dataset,
            datainfo=datainfo,
            estimation_steps=estimation_steps,
            modelfit_results=modelfit_results,
            parent_model=parent_model,
            initial_individual_estimates=initial_individual_estimates,
            filename_extension=filename_extension,
            value_type=value_type,
            description=description,
            internals=internals,
            observation_transformation=observation_transformation,
        )

    def __eq__(self, other):
        """Compare two models for equality

        Tests whether a model is equal to another model. This ignores
        implementation-specific details such as NONMEM $DATA and FILE
        pointers, or certain $TABLE printing options.

        Parameters
        ----------
        other : Model
            Other model to compare this one with

        Examples
        --------
        >>> from pharmpy.model import Model
        >>> from pharmpy.modeling import load_example_model
        >>> a = load_example_model("pheno")
        >>> a == a
        True
        >>> a == 0
        Traceback (most recent call last):
         ...
        NotImplementedError: Cannot compare Model with <class 'int'>
        >>> a == None
        Traceback (most recent call last):
         ...
        NotImplementedError: Cannot compare Model with <class 'NoneType'>
        >>> b = load_example_model("pheno")
        >>> b == a
        True
        >>> a = a.replace(name='a')
        >>> b = b.replace(name='b')
        >>> a == b
        True
        """
        if self is other:
            return True
        if not isinstance(other, Model):
            raise NotImplementedError(f'Cannot compare Model with {type(other)}')

        if self.parameters != other.parameters:
            return False
        if self.random_variables != other.random_variables:
            return False
        if self.statements != other.statements:
            return False
        if self.dependent_variables != other.dependent_variables:
            return False
        if self.observation_transformation != other.observation_transformation:
            return False
        if self.estimation_steps != other.estimation_steps:
            return False
        if self.initial_individual_estimates is None:
            if other.initial_individual_estimates is not None:
                return False
        else:
            if other.initial_individual_estimates is None:
                return False
            elif not self.initial_individual_estimates.equals(other.initial_individual_estimates):
                return False
        if self.datainfo != other.datainfo:
            return False
        if self.value_type != other.value_type:
            return False

        return True

    @cache_method
    def __hash__(self):
        dataset_hash = hash_df_runtime(self._dataset) if self._dataset is not None else None
        return hash(
            (
                self._parameters,
                self._random_variables,
                self._statements,
                self._dependent_variables,
                self._observation_transformation,
                self._estimation_steps,
                self._initial_individual_estimates,
                self._datainfo,
                dataset_hash,
                self._value_type,
            )
        )

    def to_dict(self) -> dict[str, Any]:
        if self._initial_individual_estimates is not None:
            ie = self._initial_individual_estimates.to_dict()
        else:
            ie = None
        depvars = {str(key): val for key, val in self._dependent_variables.items()}
        obstrans = {
            sympy.srepr(key): sympy.srepr(val)
            for key, val in self._observation_transformation.items()
        }
        return {
            'parameters': self._parameters.to_dict(),
            'random_variables': self._random_variables.to_dict(),
            'statements': self._statements.to_dict(),
            'estimation_steps': self._estimation_steps.to_dict(),
            'datainfo': self._datainfo.to_dict(),
            'value_type': self._value_type,
            'dependent_variables': depvars,
            'observation_transformation': obstrans,
            'initial_individual_estimates': ie,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Model:
        ie_dict = d['initial_individual_estimates']
        if ie_dict is None:
            ie = None
        else:
            ie = pd.DataFrame.from_dict(ie_dict)
        depvars = {sympy.Symbol(key): value for key, value in d['dependent_variables'].items()}
        obstrans = {
            sympy.parse_expr(key): sympy.parse_expr(val)
            for key, val in d['observation_transformation'].items()
        }
        return cls(
            parameters=Parameters.from_dict(d['parameters']),
            random_variables=RandomVariables.from_dict(d['random_variables']),
            statements=Statements.from_dict(d['statements']),
            estimation_steps=EstimationSteps.from_dict(d['estimation_steps']),
            datainfo=DataInfo.from_dict(d['datainfo']),
            value_type=d['value_type'],
            dependent_variables=frozenmapping(depvars),
            observation_transformation=frozenmapping(obstrans),
            initial_individual_estimates=ie,
        )

    def __repr__(self):
        return f'<Pharmpy model object {self.name}>'

    def _repr_html_(self):
        stat = self.statements._repr_html_()
        rvs = self.random_variables._repr_latex_()
        return f'<hr>{stat}<hr>${rvs}$<hr>{self.parameters._repr_html_()}<hr>'

    @property
    def name(self) -> str:
        """Name of the model"""
        return self._name

    @property
    def filename_extension(self) -> str:
        """Filename extension of model file"""
        return self._filename_extension

    @property
    def dependent_variables(self) -> frozenmapping[sympy.Symbol, int]:
        """The dependent variables of the model mapped to the corresponding DVIDs"""
        return self._dependent_variables

    @property
    def value_type(self) -> str:
        """The type of the model value (dependent variables)

        By default this is set to 'PREDICTION' to mean that the model outputs a prediction.
        It could optionally be set to 'LIKELIHOOD' or '-2LL' to let the model output the likelihood
        or -2*log(likelihood) of the prediction. If set to a symbol this variable can be used to
        change the type for different records. The model would then set this symbol to 0 for
        a prediction value, 1 for likelihood and 2 for -2ll.
        """
        return self._value_type

    @property
    def observation_transformation(self) -> frozenmapping[sympy.Symbol, sympy.Expr]:
        """Transformation to be applied to the observation data"""
        return self._observation_transformation

    @property
    def parameters(self) -> Parameters:
        """Definitions of population parameters

        See :class:`pharmpy.Parameters`
        """
        return self._parameters

    @property
    def random_variables(self) -> RandomVariables:
        """Definitions of random variables

        See :class:`pharmpy.RandomVariables`
        """
        return self._random_variables

    @property
    def statements(self) -> Statements:
        """Definitions of model statements

        See :class:`pharmpy.Statements`
        """
        return self._statements

    @property
    def estimation_steps(self) -> EstimationSteps:
        """Definitions of estimation steps

        See :class:`pharmpy.EstimationSteps`
        """
        return self._estimation_steps

    @property
    def datainfo(self) -> DataInfo:
        """Definitions of model statements

        See :class:`pharmpy.Statements`
        """
        return self._datainfo

    @property
    def dataset(self) -> Optional[pd.DataFrame]:
        """Dataset connected to model"""
        return self._dataset

    @property
    def initial_individual_estimates(self) -> Optional[pd.DataFrame]:
        """Initial estimates for individual parameters"""
        return self._initial_individual_estimates

    @property
    def internals(self) -> Optional[ModelInternals]:
        """Internal data for tool specific part of model"""
        return self._internals

    @property
    def modelfit_results(self):
        """Modelfit results for this model"""
        return self._modelfit_results

    @property
    def model_code(self) -> str:
        """Model type specific code"""
        d = self.to_dict()
        d['__magic__'] = "Pharmpy Model"
        d['__version__'] = pharmpy.__version__
        return json.dumps(d)

    @property
    def parent_model(self) -> Optional[str]:
        """Name of parent model"""
        return self._parent_model

    def has_same_dataset_as(self, other: Model) -> bool:
        """Check if this model has the same dataset as another model

        Parameters
        ----------
        other : Model
            Another model

        Returns
        -------
        bool
            True if both models have the same dataset
        """
        if self.dataset is None:
            if other.dataset is None:
                return True
            else:
                return False

        if other.dataset is None:
            return False

        # NOTE: Rely on duck-typing here (?)
        return self.dataset.equals(other.dataset)

    @property
    def description(self) -> str:
        """A free text description of the model"""
        return self._description

    @staticmethod
    def parse_model(path: Union[Path, str]):
        """Create a model object by parsing a model file of any supported type

        Parameters
        ----------
        path : Path or str
            Path to a model file

        Returns
        -------
        Model
            A model object
        """
        path = Path(path)
        with open(path, 'r', encoding='latin-1') as fp:
            code = fp.read()

        model_module = detect_model(code)
        model = model_module.parse_model(code, path)
        return model

    @staticmethod
    def parse_model_from_string(code: str) -> Model:
        """Create a model object by parsing a string with model code of any supported type

        Parameters
        ----------
        code : str
            Model code

        Returns
        -------
        Model
            A model object
        """
        model_module = detect_model(code)
        model = model_module.parse_model(code, None)
        return model

    def update_source(self) -> Model:
        """Update source code of the model. If any paths need to be changed or added (e.g. for a
        NONMEM model with an updated dataset) they will be replaced with DUMMYPATH"""
        return self

    def write_files(self, path: Optional[Union[Path, str]] = None, force: bool = False) -> Model:
        """Write all extra files needed for a specific external format."""
        return self


def compare_before_after_params(old, new):
    # FIXME: Move this to the right module
    before = {}
    after = {}
    for key, value in old.items():
        if new[key] != value:
            before[key] = value
            after[key] = new[key]
    return before, after


def update_datainfo(curdi: DataInfo, dataset: pd.DataFrame) -> DataInfo:
    colnames = dataset.columns
    columns = []
    for colname in colnames:
        try:
            col = curdi[colname]
        except IndexError:
            datatype = ColumnInfo.convert_pd_dtype_to_datatype(dataset.dtypes[colname].name)
            col = ColumnInfo.create(colname, datatype=datatype)
        columns.append(col)
    newdi = curdi.replace(columns=columns)

    # NOTE: Remove path if dataset has been updated
    return curdi if newdi == curdi else newdi.replace(path=None)



import json
import os
import subprocess
import sys
import uuid
import warnings
from pathlib import Path
from typing import Optional, Union

import pharmpy.model
from pharmpy.deps import pandas as pd
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.model.external.rxode import convert_model
from pharmpy.modeling import get_omegas, get_sigmas, update_inits, write_csv
from pharmpy.tools import fit
from pharmpy.tools.external.nlmixr.run import compare_models, print_step
from pharmpy.workflows import ModelEntry, default_tool_database
from pharmpy.workflows.results import ModelfitResults


def execute_model(model_entry, db):
    assert isinstance(model_entry, ModelEntry)
    model = model_entry.model

    database = db.model_database
    model = convert_model(model)
    path = Path.cwd() / f'rxode_run_{model.name}-{uuid.uuid1()}'
    model = model.replace(internals=model.internals.replace(path=path))
    meta = path / '.pharmpy'
    meta.mkdir(parents=True, exist_ok=True)
    if model.datainfo.path is not None:
        model = model.replace(datainfo=model.datainfo.replace(path=None))
    write_csv(model, path=path)
    model = model.replace(datainfo=model.datainfo.replace(path=path))

    dataname = f'{model.name}.csv'
    if sys.platform == 'win32':
        dataset_path = f"{path / dataname}".replace("\\", "\\\\")
    else:
        dataset_path = f"{path / dataname}"
    pre = f'library(rxode2)\n\nev <- read.csv("{dataset_path}")\n'

    pre += "\n"

    code = pre + model.model_code
    cg = CodeGenerator()

    dv = list(model.dependent_variables.keys())[0]
    cg.add(f"res <- as.data.frame(fit[c('id', 'time', '{dv}')])")
    cg.add("sigma <- as.data.frame(sigmas)")
    cg.add("omegas <- as.data.frame(omegas)")
    cg.add("params <- as.data.frame(fit$params)")

    if sys.platform == 'win32':
        p = f"{path / model.name}.RDATA".replace("\\", "\\\\")
    else:
        p = f"{path / model.name}.RDATA"
    cg.add(f'save(file="{p}", res, params)')

    code += f'\n{str(cg)}'

    with open(path / f'{model.name}.R', 'w') as fh:
        fh.write(code)

    from pharmpy.tools.external.nlmixr import conf

    rpath = conf.rpath / 'bin' / 'Rscript'

    newenv = os.environ
    # Reset environment variables incase started from R
    # and calling other R version.
    newenv['R_LIBS_USERS'] = ''
    newenv['R_LIBS_SITE'] = ''

    stdout = path / 'stdout'
    stderr = path / 'stderr'

    args = [str(rpath), str(path / (model.name + '.R'))]

    with open(stdout, "wb") as out, open(stderr, "wb") as err:
        result = subprocess.run(args, stdin=subprocess.DEVNULL, stderr=err, stdout=out, env=newenv)

    rdata_path = path / f'{model.name}.RDATA'

    metadata = {
        'plugin': 'nlmixr',
        'path': str(path),
    }

    plugin = {
        'rpath': str(rpath),
        'commands': [
            {
                'args': args,
                'returncode': result.returncode,
                'stdout': 'stdout',
                'stderr': 'stderr',
            }
        ],
    }

    with database.transaction(model) as txn:
        txn.store_local_file(path / f'{model.name}.R')
        txn.store_local_file(rdata_path)

        txn.store_local_file(stdout)
        txn.store_local_file(stderr)
        txn.store_local_file(path / f'{model.name}.csv')

        txn.store_local_file(model.datainfo.path)

        plugin_path = path / 'nlmixr.json'
        with open(plugin_path, 'w') as f:
            json.dump(plugin, f, indent=2)

        txn.store_local_file(plugin_path)

        txn.store_metadata(metadata)
        txn.store_modelfit_results()

    res = parse_modelfit_results(model, path)
    log = res.log if res else None
    model_entry = model_entry.attach_results(modelfit_results=res, log=log)

    return model_entry


def parse_modelfit_results(model: pharmpy.model.Model, path: Path) -> Union[None, ModelfitResults]:
    rdata_path = path / (model.name + '.RDATA')
    with warnings.catch_warnings():
        # Supress a numpy deprecation warning
        warnings.simplefilter("ignore")
        import pyreadr
    try:
        rdata = pyreadr.read_r(rdata_path)
    except (FileNotFoundError, OSError):
        return None

    dv = list(model.dependent_variables.keys())[0]
    pred = rdata["res"][["id", "time", f"{dv}"]]
    pred.rename(columns={f"{dv}": 'PRED', "id": "ID", "time": "TIME"}, inplace=True)
    pred = pred.set_index(["ID", "TIME"])

    # TODO: Extract thetas, omegas and sigmas

    predictions = pred
    predictions.index = predictions.index.set_levels(
        predictions.index.levels[0].astype("float64"), level=0
    )

    # TODO: Add more variables such as name and description and parameter estimates
    res = ModelfitResults(predictions=predictions)
    return res


def verification(
    model: pharmpy.model.Model,
    modelfit_results: Optional[ModelfitResults] = None,
    error: float = 10**-3,
    return_comp: bool = False,
    ignore_print=False,
) -> Union[bool, pd.DataFrame]:
    nonmem_model = model

    # Save results from the nonmem model
    if modelfit_results is None:
        if not ignore_print:
            print_step("Calculating NONMEM predictions... (this might take a while)")
        nonmem_res = fit(nonmem_model)
    else:
        if modelfit_results.predictions is None:
            if not ignore_print:
                print_step("Calculating NONMEM predictions... (this might take a while)")
            nonmem_res = fit(nonmem_model)
        else:
            nonmem_res = modelfit_results

    param_estimates = nonmem_res.parameter_estimates

    omega_names = get_omegas(nonmem_model).names
    for name in omega_names:
        param_estimates[name] = 0

    sigma_names = get_sigmas(model).names
    for name in sigma_names:
        param_estimates[name] = 0

    # Update the nonmem model with new estimates
    # and convert to nlmixr
    if not ignore_print:
        print_step("Converting NONMEM model to RxODE...")
    rxode_model = convert_model(update_inits(nonmem_model, param_estimates))

    # Execute the rxode model
    db = default_tool_database(toolname="comparison")
    if not ignore_print:
        print_step("Executing RxODE model... (this might take a while)")

    rxode_model_entry = ModelEntry(rxode_model, modelfit_results=None)
    rxode_model_entry = execute_model(rxode_model_entry, db)

    combined_result = compare_models(
        nonmem_model,
        nonmem_res,
        rxode_model,
        rxode_model_entry.modelfit_results,
        error=error,
        force_pred=True,
        ignore_print=ignore_print,
    )

    combined_result.to_csv(db.path / "comparison.csv", index=False)

    if not ignore_print:
        print_step("DONE")
    if return_comp is True:
        return combined_result
    else:
        if all(combined_result["PASS/FAIL"] == "PASS"):
            return True
        else:
            return False



import itertools

from pharmpy.modeling import (
    get_observations,
    remove_peripheral_compartment,
    set_initial_estimates,
    set_name,
    set_tmdd,
)


def product_dict(**kwargs):
    keys = kwargs.keys()
    for instance in itertools.product(*kwargs.values()):
        yield dict(zip(keys, instance))


def create_qss_models(model, ests, dv_types, index=1):
    # Create qss models with different initial estimates from basic pk model
    qss_base_model = set_tmdd(model, type="QSS", dv_types=dv_types)
    cmax = get_observations(model).max()
    all_inits = product_dict(
        POP_KDEG=(0.5623, 17.28), POP_R_0=(0.001 * cmax, 0.01 * cmax, 0.1 * cmax, 1 * cmax)
    )
    qss_candidate_models = [
        set_initial_estimates(set_name(qss_base_model, f"QSS{i}"), inits)
        for i, inits in enumerate(all_inits, start=index)
    ]
    if "POP_KM" in ests.keys():
        qss_candidate_models = [
            set_initial_estimates(
                model,
                {
                    "POP_KDC": ests["POP_KM"],
                },
            )
            for model in qss_candidate_models
        ]
    if 'POP_VC' in ests:
        est_vc = ests['POP_VC']
    else:
        est_vc = 0.1

    if "POP_KM" in ests.keys() and "POP_CLMM" in ests.keys():
        qss_candidate_models = [
            set_initial_estimates(
                model,
                {
                    "POP_KINT": ests["POP_KM"]
                    * ests["POP_CLMM"]
                    / (model.parameters["POP_R_0"].init * est_vc),
                },
            )
            for model in qss_candidate_models
        ]
    if "IIV_CLMM" in ests.keys():
        qss_candidate_models = [
            set_initial_estimates(model, {"IIV_R_0": ests["IIV_CLMM"]})
            for model in qss_candidate_models
        ]
    else:
        qss_candidate_models = [
            set_initial_estimates(model, {"IIV_R_0": 0.04}) for model in qss_candidate_models
        ]
    qss_candidate_models = [
        model.replace(description=f"QSS{i}")
        for i, model in enumerate(qss_candidate_models, start=index)
    ]
    return qss_candidate_models


def create_remaining_models(model, ests, num_peripherals_qss, dv_types):
    # if best qss model has fewer compartments than model, remove one compartment
    num_peripherals_model = len(model.statements.ode_system.find_peripheral_compartments())
    if num_peripherals_qss < num_peripherals_model:
        model = remove_peripheral_compartment(model)

    models = (
        create_full_models(model, ests, dv_types)
        + create_cr_models(model, ests, dv_types)
        + create_ib_models(model, ests, dv_types)
        + create_crib_models(model, ests, dv_types)
        + create_wagner_model(model, ests, dv_types)
        + create_mmapp_model(model, ests, dv_types)
    )
    return models


def create_cr_models(model, ests, dv_types):
    # Create cr models with different initial estimates from basic pk model and best qss ests
    cr_base_model = set_tmdd(model, type="CR", dv_types=dv_types)
    cr_base_model = set_initial_estimates(
        cr_base_model,
        {"POP_KINT": ests['POP_KINT'], "POP_R_0": ests['POP_R_0'], "IIV_R_0": ests['IIV_R_0']},
    )
    cr1 = set_name(cr_base_model, "structsearch_run5")
    cr1 = cr1.replace(description="CR1")
    cr1 = set_initial_estimates(cr1, ests)
    cr1 = set_initial_estimates(cr1, {"POP_KOFF": 0.5623, "POP_KON": 0.5623 / ests['POP_KDC']})
    cr2 = set_name(cr_base_model, "structsearch_run6")
    cr2 = cr2.replace(description="CR2")
    cr2 = set_initial_estimates(cr2, ests)
    cr2 = set_initial_estimates(cr2, {"POP_KOFF": 17.78, "POP_KON": 17.78 / ests['POP_KDC']})
    return [cr1, cr2]


def create_ib_models(model, ests, dv_types):
    # Create ib models with different initial estimates from basic pk model and best qss ests
    ib_base_model = set_tmdd(model, type="IB", dv_types=dv_types)
    ib_base_model = set_initial_estimates(
        ib_base_model,
        {
            "POP_KINT": ests['POP_KINT'],
            "POP_R_0": ests['POP_R_0'],
            "POP_KDEG": ests['POP_KDEG'],
            "IIV_R_0": ests['IIV_R_0'],
        },
    )
    ib1 = set_name(ib_base_model, "structsearch_run7")
    ib1 = ib1.replace(description="IB1")
    ib1 = set_initial_estimates(ib1, ests)
    ib1 = set_initial_estimates(ib1, {"POP_KON": 0.5623 / ests['POP_KDC']})
    ib2 = set_name(ib_base_model, "structsearch_run8")
    ib2 = ib2.replace(description="IB2")
    ib2 = set_initial_estimates(ib2, ests)
    ib2 = set_initial_estimates(ib2, {"POP_KON": 17.78 / ests['POP_KDC']})
    return [ib1, ib2]


def create_crib_models(model, ests, dv_types):
    # Create crib models with different initial estimates from basic pk model and best qss ests
    crib_base_model = set_tmdd(model, type="CRIB", dv_types=dv_types)
    crib_base_model = set_initial_estimates(
        crib_base_model,
        {
            "POP_KINT": ests['POP_KINT'],
            "POP_R_0": ests['POP_R_0'],
            "IIV_R_0": ests['IIV_R_0'],
            "POP_KDEG": ests['POP_KDEG'],
        },
    )
    crib1 = set_name(crib_base_model, "structsearch_run9")
    crib1 = crib1.replace(description="CR+IB1")
    crib1 = set_initial_estimates(crib1, ests)
    crib1 = set_initial_estimates(crib1, {"POP_KON": 0.5623 / ests['POP_KDC']})
    crib2 = set_name(crib_base_model, "structsearch_run10")
    crib2 = crib2.replace(description="CR+IB2")
    crib2 = set_initial_estimates(crib2, ests)
    crib2 = set_initial_estimates(crib2, {"POP_KON": 17.78 / ests['POP_KDC']})
    return [crib1, crib2]


def create_full_models(model, ests, dv_types):
    # Create full models with different initial estimates from basic pk model and best qss ests
    full_base_model = set_tmdd(model, type="FULL", dv_types=dv_types)
    full_base_model = set_initial_estimates(full_base_model, ests)
    full_base_model = set_initial_estimates(
        full_base_model,
        {
            "POP_KINT": ests['POP_KINT'],
            "POP_R_0": ests['POP_R_0'],
            "IIV_R_0": ests['IIV_R_0'],
            "POP_KDEG": ests['POP_KDEG'],
        },
    )
    candidates = [
        set_initial_estimates(full_base_model, {'POP_KOFF': koff}) for koff in (0.1, 1, 10, 100)
    ]
    candidates = [
        set_initial_estimates(
            model, {'POP_KON': model.parameters['POP_KOFF'].init / ests['POP_KDC']}
        )
        for model in candidates
    ]
    candidates = [set_name(model, f"structsearch_run{i}") for i, model in enumerate(candidates, 1)]
    candidates = [m.replace(description=f"FULL{i}") for i, m in enumerate(candidates, 1)]
    return candidates


def create_wagner_model(model, ests, dv_types):
    wagner = set_tmdd(model, type="WAGNER", dv_types=dv_types)
    wagner = set_name(wagner, "structsearch_run11")
    wagner = wagner.replace(description="WAGNER")
    wagner = set_initial_estimates(
        wagner,
        {
            "POP_KINT": ests['POP_KINT'],
            "POP_R_0": ests['POP_R_0'],
            "IIV_R_0": ests['IIV_R_0'],
            "POP_KM": ests['POP_KDC'],
        },
    )
    return [wagner]


def create_mmapp_model(model, ests, dv_types):
    mmapp = set_tmdd(model, type="MMAPP", dv_types=dv_types)
    mmapp = set_name(mmapp, "structsearch_run12")
    mmapp = mmapp.replace(description="MMAPP")
    mmapp = set_initial_estimates(
        mmapp,
        {
            "POP_KINT": ests['POP_KINT'],
            "POP_R_0": ests['POP_R_0'],
            "IIV_R_0": ests['IIV_R_0'],
            "POP_KDEG": ests['POP_KDEG'],
            "POP_KM": ests['POP_KDC'],
        },
    )
    return [mmapp]



"""Utils for managing detecting and parsing of external model types."""

from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules


def detect_model(src):
    """Detects appropriate implementation from a source object
    Return an external model module
    """

    plugins = _load_external_modules()
    detected_modules = []
    for module in plugins:
        if hasattr(module, 'detect_model'):
            is_module = module.detect_model(src)
            if is_module:
                detected_modules.append(module)

    if len(detected_modules) == 0:
        raise TypeError(f"No support for model {src}")
    elif len(detected_modules) > 1:
        raise TypeError(f"More than one external model module supports model {src}")
    else:
        return detected_modules[0]


def _load_external_modules():
    """Find and import all available external modules"""
    path = Path(__file__).resolve().parent

    # str on path to workaround https://bugs.python.org/issue44061
    path = str(path)

    return [
        import_module(modname)
        for _, modname, ispkg in iter_modules([path], 'pharmpy.model.external.')
        if ispkg
    ]



import pytest

import pharmpy.config as config
from pharmpy.model import data


def test_config_item():
    class TestClass:
        item1 = config.ConfigItem(23, "A single number")

    obj = TestClass()
    assert obj.item1 == 23
    obj.item1 = 91
    assert obj.item1 == 91
    with pytest.raises(TypeError):
        obj.item1 = "A"


def test_data_config():
    assert data.conf.na_values == [-99]



import pytest

from pharmpy.modeling import fix_parameters
from pharmpy.modeling.help_functions import _as_integer, _get_etas


def test_get_etas(pheno_path, testdata, load_model_for_test):
    model = load_model_for_test(pheno_path)

    etas = _get_etas(model, ['ETA_1'])
    assert len(etas) == 1

    etas = _get_etas(model, ['ETA_1', 'CL'], include_symbols=True)
    assert len(etas) == 1

    etas = _get_etas(model, ['ETA_1', 'V'], include_symbols=True)
    assert len(etas) == 2

    with pytest.raises(KeyError):
        _get_etas(model, ['ETA_23'])

    model = load_model_for_test(testdata / 'nonmem' / 'pheno_block.mod')
    rvs = _get_etas(model, None)
    assert rvs[0] == 'ETA_CL'

    fix_parameters(model, ['OMEGA_1_1'])
    rvs = _get_etas(model, None)
    assert rvs[0] == 'ETA_CL'

    model = load_model_for_test(testdata / 'nonmem' / 'pheno_block.mod')
    new_eta = model.random_variables['ETA_CL'].replace(level='IOV')
    model = model.replace(random_variables=new_eta + model.random_variables[1:])
    rvs = _get_etas(model, None)
    assert rvs[0] == 'ETA_V'


def test_as_integer():
    n = _as_integer(1)
    assert isinstance(n, int) and n == 1
    n = _as_integer(1.0)
    assert isinstance(n, int) and n == 1
    with pytest.raises(TypeError):
        _as_integer(1.5)



import warnings
from dataclasses import dataclass, replace
from itertools import product
from pathlib import Path
from typing import Any, Optional

from pharmpy.deps import altair as alt
from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.deps import symengine, sympy
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.math import (
    conditional_joint_normal,
    conditional_joint_normal_lambda,
    is_posdef,
)
from pharmpy.model import Model
from pharmpy.modeling import (
    calculate_individual_shrinkage,
    create_rng,
    get_baselines,
    get_covariate_baselines,
    is_linearized,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    set_covariates,
)
from pharmpy.tools import read_modelfit_results
from pharmpy.workflows import Results, ToolDatabase


@dataclass(frozen=True)
class FREMResults(Results):
    """FREM Results class

    What follows is a description on how the FREM results are stored.
    See :py:mod:`pharmpy.tools.frem` for a description of how the results are calculated.

    .. attribute:: covariate_baselines

        DataFrame with the baseline covariate values used in the analysis for each indiviual.
        Index is ID. One column for each covariate.

        .. code-block::

                  WGT  APGR
            ID
            1.0   1.4   7.0
            2.0   1.5   9.0
            3.0   1.5   6.0
            4.0   0.9   6.0
            5.0   1.4   7.0
            6.0   1.2   5.0
            7.0   1.0   5.0
            8.0   1.2   7.0
            9.0   1.4   8.0
            ...   ...   ...


    .. attribute:: covariate_effects

        DataFrame with covariate effects. Index is parameter, covariate and condition where
        condition can be either 5th or 95th. Columns are p5, mean and p95 for the 5th
        percentile of the effect, the estimated effect and the 95th percentile of the effect
        respectively. Effect sizes are in fractions. Example:

        .. code-block::

                                                 p5      mean       p95
            parameter covariate condition
            ETA(1)    WGT       5th        0.901869  0.972920  1.051093
                                95th       0.903849  1.064605  1.233111
                      APGR      5th        1.084238  1.248766  1.413923
                                95th       0.848297  0.901919  0.962312
            ETA(2)    WGT       5th        0.942450  1.004400  1.068390
                                95th       0.874409  0.995763  1.127777
                      APGR      5th        0.832008  0.924457  1.021307
                                95th       0.990035  1.039444  1.091288

    .. attribute:: covariate_effects_plot

        Plot of the covariate effects generated by :py:meth:`plot_covariate_effects`

    .. attribute:: covariate_statistics

        DataFrame with summary statistics of the covariate baselines. Index is covariates.
        Columns are p5 with 95th percentile, mean, p95 with 95th percentile, stdev, ref with
        the used reference value, categorical and other with the non-reference value for
        categorical covariates. Example:

        .. code-block

                                    p5      mean  p95     stdev       ref  categorical  other
                        covariate
                        WGT        0.7  1.525424  3.2  0.704565  1.525424        False    NaN
                        APGR       1.0  6.423729  9.0  2.237636  6.423729        False    NaN

    .. attribute:: individual_effects

        DataFrame with individual covariate effects. Index is ID and parameter. Columns
        are observed, p5 and p95 for the observed individual effect, 5th and 95th
        percentiles of the estimated individual effects respectively. Example:

        .. code-block::

                            observed        p5       p95
            ID   parameter
            1.0  ETA(1)     0.973309  0.946282  0.982391
                 ETA(2)     1.009411  0.995276  1.024539
            2.0  ETA(1)     0.911492  0.832168  0.941760
                 ETA(2)     1.036200  0.990081  1.091438
            3.0  ETA(1)     1.013772  1.007555  1.028463
            ...                  ...       ...       ...
            57.0 ETA(2)     0.987500  0.942709  1.031111
            58.0 ETA(1)     0.939409  0.883782  0.956792
                 ETA(2)     1.023321  0.993543  1.057408
            59.0 ETA(1)     0.992578  0.952785  1.027261
                 ETA(2)     0.999220  0.968931  1.033819

    .. attribute:: individual_effects_plot

        Plot of the individual effects generated by :py:meth:`plot_individual_effects`

    .. attribute:: unexplained_variability

        DataFrame with remaining unexplained variability. Index is parameter and covariate.
        Covariate is none for no covariates and all for all covariates. Example:

        .. code-block::

                                 sd_observed    sd_5th   sd_95th
            parameter covariate
            ETA(1)    none          0.195327  0.221382  0.298465
                      WGT           0.194527  0.218385  0.292261
                      APGR          0.182497  0.202192  0.279766
                      all           0.178956  0.197282  0.268090
            ETA(2)    none          0.158316  0.152510  0.210044
                      WGT           0.158313  0.148041  0.207276
                      APGR          0.155757  0.149037  0.203477
                      all           0.155551  0.144767  0.201658

    .. attribute:: unexplained_variability_plot

        Plot of the unexplained variability generated by :py:meth:`plot_unexplained_variability`

    .. attribute:: parameter_inits_and_estimates

        Initial parameter estimates and estimates after fitting for all intermediate models and
        the final model.

    .. attribute:: base_parameter_change

        The relative change in parameter estimates from base model to the FREM model.

    .. attribute:: covariate_estimates

        Model estimates of covariate statistics

    .. attribute:: parameter_variability

        Conditioned parameter variability

    .. attribute:: coefficients

        Parameter covariate coefficients. Calculated one at a time or all together.

    """

    rst_path = Path(__file__).resolve().parent / 'report.rst'

    coefficients: Optional[Any] = None
    parameter_variability: Optional[Any] = None
    covariate_effects: Optional[Any] = None
    individual_effects: Optional[Any] = None
    unexplained_variability: Optional[Any] = None
    covariate_statistics: Optional[Any] = None
    covariate_effects_plot: Optional[Any] = None
    individual_effects_plot: Optional[Any] = None
    unexplained_variability_plot: Optional[Any] = None
    covariate_baselines: Optional[Any] = None
    parameter_inits_and_estimates: Optional[Any] = None
    base_parameter_change: Optional[Any] = None
    estimated_covariates: Optional[Any] = None
    ofv: Optional[Any] = None
    tool_database: Optional[ToolDatabase] = None


def plot_covariate_effects(res):
    """Plot covariate effects"""
    ce = (res.covariate_effects - 1) * 100
    cov_stats = pd.melt(
        res.covariate_statistics.reset_index(),
        var_name='condition',
        id_vars=['covariate'],
        value_vars=['p5', 'p95', 'other'],
    )

    cov_stats = cov_stats.replace({'p5': '5th', 'p95': '95th'}).set_index(
        ['covariate', 'condition']
    )

    ce = ce.join(cov_stats, how='inner')

    # The left join reorders the index, pandas bug #34133
    ce = ce.reorder_levels(['parameter', 'covariate', 'condition'])

    param_names = list(ce.index.get_level_values('parameter').unique())
    plots = []

    for parameter in param_names:
        df = ce.xs(parameter, level=0)
        df = df.reset_index()

        error_bars = (
            alt.Chart(df)
            .mark_errorbar(ticks=True)
            .encode(
                x=alt.X('p5:Q', title='Effect size in percent', scale=alt.Scale(zero=False)),
                x2=alt.X2('p95:Q'),
                y=alt.Y('condition:N', title=None),
            )
        )

        rule = (
            alt.Chart(df)
            .mark_rule(strokeDash=[10, 4], color='gray')
            .encode(x=alt.X('xzero:Q'))
            .transform_calculate(xzero="0")
        )

        points = (
            alt.Chart(df)
            .mark_point(filled=True, color='black')
            .encode(
                x=alt.X('mean:Q'),
                y=alt.Y('condition:N'),
            )
        )

        text = (
            alt.Chart(df)
            .mark_text(dy=-15, color="red")
            .encode(x=alt.X("mean:Q"), y=alt.Y("condition:N"), text=alt.Text("value:Q"))
        )

        plot = (
            alt.layer(error_bars, rule, points, text, data=df, width=700, height=100)
            .facet(columns=1.0, row=alt.Facet('covariate:N', title=None), title=f'{parameter}')
            .resolve_scale(y='independent')
        )

        plots.append(plot)

    v = alt.vconcat(*plots).resolve_scale(x='shared')
    return v


def plot_individual_effects(res):
    """Plot individual effects"""
    covs = res.covariate_baselines
    ie = res.individual_effects.join(covs)
    param_names = list(ie.index.get_level_values('parameter').unique())
    ie = (ie - 1) * 100
    ie = ie.sort_values(by=['observed'])

    plots = []

    for parameter in param_names:
        df = ie.xs(parameter, level=1)

        id_order = list(df.index)
        id_order = [str(int(x)) for x in id_order]

        if len(df) > 20:
            id_order[10] = '...'

        df = df.reset_index()
        df['ID'] = df['ID'].astype(int).astype(str)

        error_bars = (
            alt.Chart(df)
            .mark_errorbar(ticks=True)
            .encode(
                x=alt.X('p5:Q', title='Effect size in percent', scale=alt.Scale(zero=False)),
                x2=alt.X2('p95:Q'),
                y=alt.Y('ID:N', title='ID', sort=id_order),
                tooltip=['ID', 'p5', 'observed', 'p95'] + list(covs.columns),
            )
        )

        rule = (
            alt.Chart(df)
            .mark_rule(strokeDash=[10, 2], color='gray')
            .encode(x=alt.X('xzero:Q'))
            .transform_calculate(xzero="0")
        )

        points = (
            alt.Chart(df)
            .mark_point(size=40, filled=True, color='black')
            .encode(
                x=alt.X('observed:Q'),
                y=alt.Y('ID:N', sort=id_order),
            )
        )

        plot = alt.layer(
            points,
            error_bars,
            rule,
            data=df,
            width=700,
            title=f'Individuals for parameter {parameter}',
        )
        if len(df) > 20:
            plot = (
                plot.transform_window(
                    sort=[alt.SortField('observed', order='ascending')],
                    rank='row_number(observed)',
                )
                .transform_window(
                    sort=[alt.SortField('observed', order='descending')],
                    nrank='row_number(observed)',
                )
                .transform_filter('datum.rank <= 10 | datum.nrank <= 11')
                .transform_calculate(
                    ID="datum.nrank == 11 ? '...' : datum.ID",
                    p5="datum.nrank == 11 ? '...' : datum.p5",
                    p95="datum.nrank == 11 ? '...' : datum.p95",
                    observed="datum.nrank == 11 ? '...' : datum.observed",
                )
            )
        plots.append(plot)

    v = alt.vconcat(*plots).resolve_scale(x='shared')
    return v


def plot_unexplained_variability(res):
    """Plot unexplained variability"""
    uv = res.unexplained_variability
    param_names = list(uv.index.get_level_values('parameter').unique())
    cov_order = list(uv.index.get_level_values('covariate').unique())
    plots = []

    for parameter in param_names:
        df = uv.xs(parameter, level=0)
        df = df.reset_index()

        error_bars = (
            alt.Chart(df)
            .mark_errorbar(ticks=True)
            .encode(
                x=alt.X(
                    'sd_5th:Q',
                    title='SD of unexplained variability',
                    scale=alt.Scale(zero=False),
                ),
                x2=alt.X2('sd_95th:Q'),
                y=alt.Y('covariate:N', title='covariate', sort=cov_order),
                tooltip=['sd_5th', 'sd_observed', 'sd_95th', 'covariate'],
            )
        )

        rule = (
            alt.Chart(df)
            .mark_rule(strokeDash=[10, 2], color='gray')
            .encode(x=alt.X('xzero:Q'))
            .transform_calculate(xzero="0")
        )

        points = (
            alt.Chart(df)
            .mark_point(size=40, filled=True, color='black')
            .encode(
                x=alt.X('sd_observed:Q'),
                y=alt.Y('covariate:N', sort=cov_order),
            )
        )

        plot = alt.layer(
            points,
            error_bars,
            rule,
            data=df,
            width=700,
            title=f'Unexplained variability on {parameter}',
        )

        plots.append(plot)

    v = alt.vconcat(*plots).resolve_scale(x='shared')
    return v


def calculate_results(
    frem_model,
    frem_model_results,
    continuous,
    categorical,
    method=None,
    intermediate_models=None,
    intermediate_models_res=None,
    seed=None,
    **kwargs,
):
    """Calculate FREM results

    :param method: Either 'cov_sampling' or 'bipp'
    """
    if intermediate_models is None:
        intermediate_models = []
        intermediate_models_res = []

    if method is None or method == 'cov_sampling':
        try:
            res = calculate_results_using_cov_sampling(
                frem_model, frem_model_results, continuous, categorical, seed=seed, **kwargs
            )
        except AttributeError:
            # Fallback to bipp
            res = calculate_results_using_bipp(
                frem_model, frem_model_results, continuous, categorical, seed=seed
            )
    elif method == 'bipp':
        res = calculate_results_using_bipp(
            frem_model, frem_model_results, continuous, categorical, seed=seed
        )
    else:
        raise ValueError(f'Unknown frem postprocessing method {method}')
    mod_names = []
    mod_ofvs = []
    if intermediate_models:
        for intmod, intmod_res in zip(intermediate_models, intermediate_models_res):
            if intmod_res is not None:
                mod_ofvs.append(intmod_res.ofv)
                mod_names.append(intmod.name)
    mod_ofvs.append(frem_model_results.ofv)
    mod_names.append(frem_model.name)
    ofv = pd.DataFrame({'ofv': mod_ofvs}, index=mod_names)
    ofv.index.name = 'model_name'
    estimates = parameter_inits_and_estimates(
        frem_model, frem_model_results, intermediate_models, intermediate_models_res
    )
    res = replace(res, ofv=ofv, parameter_inits_and_estimates=estimates)

    if intermediate_models:
        ser = base_vs_frem_model(frem_model_results, intermediate_models_res[0])
        res = replace(res, base_parameter_change=ser)

    estimated_covbase = _calculate_covariate_baselines(
        frem_model, frem_model_results, continuous + categorical
    )
    mean = estimated_covbase.mean()
    stdev = estimated_covbase.std()
    estcovs = pd.DataFrame({'mean': mean, 'stdev': stdev})

    res = replace(res, estimated_covariates=estcovs)

    return replace(
        res,
        covariate_effects_plot=plot_covariate_effects(res),
        individual_effects_plot=plot_individual_effects(res),
        unexplained_variability_plot=plot_unexplained_variability(res),
    )


def base_vs_frem_model(frem_model_res, model_1_res):
    base_ests = model_1_res.parameter_estimates
    final_ests = frem_model_res.parameter_estimates
    ser = pd.Series(dtype=np.float64, name='relative_change')
    for param in base_ests.keys():
        if param in final_ests:
            ser[param] = (final_ests[param] - base_ests[param]) / abs(base_ests[param]) * 100
    return ser


def parameter_inits_and_estimates(
    frem_model, frem_model_results, intermediate_models, intermediate_models_res
) -> pd.DataFrame:
    model_names = []
    df = pd.DataFrame()

    for model, res in zip(intermediate_models, intermediate_models_res):
        df = pd.concat(
            [
                df,
                pd.Series(model.parameters.nonfixed.inits),
                res.parameter_estimates,
            ],
            ignore_index=True,
            axis=1,
        )
        model_names.append(model.name)

    df = pd.concat(
        [
            df,
            pd.Series(frem_model.parameters.nonfixed.inits),
            frem_model_results.parameter_estimates,
        ],
        ignore_index=True,
        axis=1,
    )
    df = df.T
    df = df.reindex(columns=frem_model.parameters.nonfixed.inits.keys())
    model_names.append(frem_model.name)
    index = pd.MultiIndex.from_product([model_names, ['init', 'estimate']], names=['model', 'type'])
    df.index = index
    return df


def calculate_results_using_cov_sampling(
    frem_model,
    frem_model_results,
    continuous,
    categorical,
    cov_model=None,
    force_posdef_samples=500,
    force_posdef_covmatrix=False,
    samples=1000,
    rescale=True,
    seed=None,
):
    """Calculate the FREM results using covariance matrix for uncertainty

    :param cov_model: Take the parameter uncertainty covariance matrix from this model
                      instead of the frem model.
    :param force_posdef_samples: The number of sampling tries before stopping to use
                                 rejection sampling and instead starting to shift values so
                                 that the frem matrix becomes positive definite. Set to 0 to
                                 always force positive definiteness.
    :param force_posdef_covmatrix: Set to force the covariance matrix of the frem movdel or
                                   the cov model to be positive definite. Default is to raise
                                   in this case.
    :param samples: The number of parameter vector samples to use.
    """
    if cov_model is not None:
        uncertainty_results = cov_model.modelfit_results
    else:
        uncertainty_results = frem_model_results

    dist = frem_model.random_variables.iiv[-1]
    sigma_symb = dist.variance

    parameters = [
        s
        for s in frem_model_results.parameter_estimates.index
        if sympy.Symbol(s) in sigma_symb.free_symbols
    ]
    parvecs = sample_parameters_from_covariance_matrix(
        frem_model,
        uncertainty_results.parameter_estimates[parameters],
        uncertainty_results.covariance_matrix,
        force_posdef_samples=force_posdef_samples,
        force_posdef_covmatrix=force_posdef_covmatrix,
        n=samples,
        seed=seed,
    )
    res = calculate_results_from_samples(
        frem_model, frem_model_results, continuous, categorical, parvecs, rescale=rescale
    )
    return res


def calculate_results_from_samples(
    frem_model, frem_model_results, continuous, categorical, parvecs, rescale=True
):
    """Calculate the FREM results given samples of parameter estimates"""
    n = len(parvecs)
    dist = frem_model.random_variables.iiv[-1]
    rvs = list(dist.names)
    sigma_symb = dist.variance
    parameters = [
        s
        for s in frem_model_results.parameter_estimates.index
        if sympy.Symbol(s) in sigma_symb.free_symbols
    ]
    parvecs.loc['estimates'] = frem_model_results.parameter_estimates.loc[parameters]

    covariates = continuous + categorical
    frem_model = set_covariates(frem_model, covariates)
    covariate_baselines = get_covariate_baselines(frem_model)
    covariate_baselines = covariate_baselines[covariates]
    cov_means = covariate_baselines.mean()
    cov_modes = covariate_baselines.mode().iloc[0]  # Select first mode if more than one
    cov_others = pd.Series(index=cov_modes[categorical].index, dtype=np.float64)
    cov_stdevs = covariate_baselines.std()
    for _, row in covariate_baselines.iterrows():
        for cov in categorical:
            if row[cov] != cov_modes[cov]:
                cov_others[cov] = row[cov]
        if not cov_others.isna().values.any():
            break

    cov_refs = pd.concat((cov_means[continuous], cov_modes[categorical]))
    cov_5th = covariate_baselines.quantile(0.05, interpolation='lower')
    cov_95th = covariate_baselines.quantile(0.95, interpolation='higher')
    is_categorical = cov_refs.index.isin(categorical)

    covariate_statistics = pd.DataFrame(
        {
            'p5': cov_5th,
            'mean': cov_means,
            'p95': cov_95th,
            'stdev': cov_stdevs,
            'ref': cov_refs,
            'categorical': is_categorical,
            'other': cov_others,
        },
        index=covariates,
    )

    covariate_statistics.index.name = 'covariate'

    ncovs = len(covariates)
    npars = sigma_symb.rows - ncovs
    param_names = get_params(frem_model, rvs, npars)
    nids = len(covariate_baselines)
    param_indices = list(range(npars))
    scaling = np.diag(np.concatenate((np.ones(npars), cov_stdevs.values))) if rescale else None

    mu_bars_given_5th = np.empty((n, ncovs, npars))
    mu_bars_given_95th = np.empty((n, ncovs, npars))
    mu_id_bars = np.empty((n, nids, npars))
    original_id_bar = np.empty((nids, npars))
    variability = np.empty((n, ncovs + 2, npars))  # none, cov1, cov2, ..., all
    original_variability = np.empty((ncovs + 2, npars))
    coefficients_index = pd.MultiIndex.from_product(
        [['all', 'each'], param_names], names=['condition', 'parameter']
    )
    coefficients = pd.DataFrame(index=coefficients_index, columns=covariates, dtype=np.float64)
    parameter_variability = []

    # Switch to symengine for speed
    # Could also assume order of parameters, but not much gain
    sigma_symb = symengine.sympify(sigma_symb)
    parvecs.columns = [symengine.Symbol(colname) for colname in parvecs.columns]

    estimated_covbase = _calculate_covariate_baselines(frem_model, frem_model_results, covariates)
    covbase = estimated_covbase.to_numpy()

    parameter_variability_all = None

    for sample_no, params in parvecs.iterrows():
        sigma = sigma_symb.subs(dict(params))
        sigma = np.array(sigma).astype(np.float64)
        if scaling is not None:
            sigma = scaling @ sigma @ scaling
        if sample_no != 'estimates':
            variability[sample_no, 0, :] = np.diag(sigma)[:npars]
        else:
            original_variability[0, :] = np.diag(sigma)[:npars]
            # Coefficients conditioned on all parameters
            # Sigma_12 * Sigma_22^-1
            coeffs_all = sigma[0:npars, npars:] @ np.linalg.inv(sigma[npars:, npars:])
            coefficients.loc['all'] = coeffs_all
        for i, cov in enumerate(covariates):
            indices = param_indices + [i + npars]
            cov_sigma = sigma[indices][:, indices]
            cov_mu = np.array([0] * npars + [cov_refs[cov]])
            if cov in categorical:
                first_reference = cov_others[cov]
            else:
                first_reference = cov_5th[cov]
            mu_bar_given_5th_cov, sigma_bar = conditional_joint_normal(
                cov_mu, cov_sigma, np.array([first_reference])
            )
            if sample_no != 'estimates':
                mu_bar_given_95th_cov, _ = conditional_joint_normal(
                    cov_mu, cov_sigma, np.array([cov_95th[cov]])
                )
                mu_bars_given_5th[sample_no, i, :] = mu_bar_given_5th_cov
                mu_bars_given_95th[sample_no, i, :] = mu_bar_given_95th_cov
                variability[sample_no, i + 1, :] = np.diag(sigma_bar)
            else:
                original_variability[i + 1, :] = np.diag(sigma_bar)
                parameter_variability.append(sigma_bar)
                # Calculate coefficients
                # Cov(Par, covariate) / Var(covariate)
                for parind, parname in enumerate(param_names):
                    coefficients[cov]['each', parname] = cov_sigma[parind][-1] / cov_sigma[-1][-1]

        id_mu = np.array([0] * npars + list(cov_refs))
        cjn = conditional_joint_normal_lambda(id_mu, sigma, npars)
        for i in range(len(estimated_covbase)):
            row = covbase[i, :]
            assert npars + len(row) == len(id_mu)
            mu_id_bar, sigma_id_bar = cjn(row)
            if sample_no != 'estimates':
                mu_id_bars[sample_no, i, :] = mu_id_bar
                variability[sample_no, -1, :] = np.diag(sigma_id_bar)
            else:
                original_id_bar[i, :] = mu_id_bar
                original_variability[ncovs + 1, :] = np.diag(sigma_id_bar)
                parameter_variability_all = sigma_id_bar

    # Create covariate effects table
    mu_bars_given_5th = np.exp(mu_bars_given_5th)
    mu_bars_given_95th = np.exp(mu_bars_given_95th)

    means_5th = np.mean(mu_bars_given_5th, axis=0)
    means_95th = np.mean(mu_bars_given_95th, axis=0)
    q5_5th = np.quantile(mu_bars_given_5th, 0.05, axis=0)
    q5_95th = np.quantile(mu_bars_given_95th, 0.05, axis=0)
    q95_5th = np.quantile(mu_bars_given_5th, 0.95, axis=0)
    q95_95th = np.quantile(mu_bars_given_95th, 0.95, axis=0)

    param_col = []
    cov_col = []
    cond_col = []
    p5_col = []
    mean_col = []
    p95_col = []

    for param, cov in product(range(npars), range(ncovs)):
        if covariates[cov] in categorical:
            param_col.append(param_names[param])
            cov_col.append(covariates[cov])
            cond_col.append('other')
            p5_col.append(q5_5th[cov, param])
            mean_col.append(means_5th[cov, param])
            p95_col.append(q95_5th[cov, param])
        else:
            param_col.append(param_names[param])
            cov_col.append(covariates[cov])
            cond_col.append('5th')
            p5_col.append(q5_5th[cov, param])
            mean_col.append(means_5th[cov, param])
            p95_col.append(q95_5th[cov, param])

            param_col.append(param_names[param])
            cov_col.append(covariates[cov])
            cond_col.append('95th')
            p5_col.append(q5_95th[cov, param])
            mean_col.append(means_95th[cov, param])
            p95_col.append(q95_95th[cov, param])

    df = pd.DataFrame(
        {
            'parameter': param_col,
            'covariate': cov_col,
            'condition': cond_col,
            'p5': p5_col,
            'mean': mean_col,
            'p95': p95_col,
        }
    )
    df.set_index(['parameter', 'covariate', 'condition'], inplace=True)
    covariate_effects = df

    # Create id table
    mu_id_bars = np.exp(mu_id_bars)
    original_id_bar = np.exp(original_id_bar)

    with warnings.catch_warnings():
        # Needed because warnings emitted on Python 3.10
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        with np.testing.suppress_warnings() as sup:  # Would warn in case of missing covariates
            sup.filter(RuntimeWarning, "All-NaN slice encountered")
            id_5th = np.nanquantile(mu_id_bars, 0.05, axis=0)
            id_95th = np.nanquantile(mu_id_bars, 0.95, axis=0)

    param_col = []
    obs_col = []
    p5_col = []
    p95_col = []
    ind_col = []
    for curid, param in product(range(nids), range(npars)):
        param_col.append(param_names[param])
        obs_col.append(original_id_bar[curid, param])
        p5_col.append(id_5th[curid, param])
        p95_col.append(id_95th[curid, param])
        ind_col.append(covariate_baselines.index[curid])
    df = pd.DataFrame(
        {'parameter': param_col, 'observed': obs_col, 'p5': p5_col, 'p95': p95_col}, index=ind_col
    )
    df.index.name = 'ID'
    df = df.set_index('parameter', append=True)
    individual_effects = df

    # Create unexplained variability table
    sd_5th = np.sqrt(np.nanquantile(variability, 0.05, axis=0))
    sd_95th = np.sqrt(np.nanquantile(variability, 0.95, axis=0))
    original_sd = np.sqrt(original_variability)

    param_col = []
    cov_col = []
    sdobs_col = []
    sd5th_col = []
    sd95th_col = []
    for par, cond in product(range(npars), range(ncovs + 2)):
        if cond == 0:
            condition = 'none'
        elif cond == ncovs + 1:
            condition = 'all'
        else:
            condition = covariates[cond - 1]
        param_col.append(param_names[par])
        cov_col.append(condition)
        sdobs_col.append(original_sd[cond, par])
        sd5th_col.append(sd_5th[cond, par])
        sd95th_col.append(sd_95th[cond, par])
    df = pd.DataFrame(
        {
            'parameter': param_col,
            'covariate': cov_col,
            'sd_observed': sdobs_col,
            'sd_5th': sd5th_col,
            'sd_95th': sd95th_col,
        }
    )
    df = df.set_index(['parameter', 'covariate'])
    unexplained_variability = df

    # Create frem_parameter_variability
    index = pd.MultiIndex.from_product(
        [['all'] + covariates, param_names], names=['condition', 'parameter']
    )
    df = pd.DataFrame(index=index)
    indices = range(len(param_names))
    if parameter_variability_all is not None:
        for i, j in product(indices, repeat=2):
            df.loc[('all', param_names[i]), param_names[j]] = parameter_variability_all[i][j]
    for (k, name), i, j in product(enumerate(covariates), indices, indices):
        df.loc[(name, param_names[i]), param_names[j]] = parameter_variability[k][i][j]
    parameter_variability = df

    return FREMResults(
        covariate_statistics=covariate_statistics,
        coefficients=coefficients,
        covariate_effects=covariate_effects,
        individual_effects=individual_effects,
        unexplained_variability=unexplained_variability,
        covariate_baselines=covariate_baselines,
        parameter_variability=parameter_variability,
    )


def get_params(frem_model, rvs, npars):
    param_names = rvs[:npars]
    sset = reversed(frem_model.statements.before_odes) + frem_model.statements.error
    symbs = []

    for p in param_names:
        statement = [s for s in sset if sympy.Symbol(p) in s.rhs_symbols][0]
        if str(statement.expression) == p:
            statement = [s for s in sset if statement.symbol in s.rhs_symbols][0]
        symbs.append(statement.symbol.name)

    duplicates = set([e for e in symbs if symbs.count(e) > 1])

    for i, s in enumerate(symbs):
        if s in duplicates:
            symbs[i] = rename_duplicate(symbs, s)

    # Special case for linearized models to not get "ERR_n"
    if is_linearized(frem_model):
        nums = [int(s[3:]) for s in symbs]
        symbs = [f'ETA_{n}' for n in nums]

    return symbs


def rename_duplicate(params, stem):
    i = 1
    while True:
        candidate = f'{stem}({i})'
        if candidate not in params:
            return candidate
        i += 1


def _calculate_covariate_baselines(model, res, covariates):
    exprs = [
        ass.expression.args[0][0]
        for ass in model.statements
        if sympy.Symbol('FREMTYPE') in ass.free_symbols and ass.symbol.name == 'IPRED'
    ]
    exprs = [
        subs(
            subs(model.statements.before_odes.full_expression(expr), res.parameter_estimates),
            model.parameters.inits,
        )
        for expr in exprs
    ]

    def fn(row):
        return [np.float64(subs(expr, dict(row))) for expr in exprs]

    df = res.individual_estimates.apply(fn, axis=1, result_type='expand')
    df.columns = covariates
    return df


def calculate_results_using_bipp(
    frem_model, frem_model_results, continuous, categorical, rescale=True, samples=2000, seed=None
):
    """Estimate a covariance matrix for the frem model using the BIPP method

    Bootstrap on the individual parameter posteriors
    Only the individual estimates, individual unvertainties and the parameter estimates
    are needed.

    """
    rng = create_rng(seed)
    assert rng is not None
    dist = frem_model.random_variables.iiv[-1]
    etas = list(dist.names)
    pool = sample_individual_estimates(
        frem_model,
        frem_model_results.individual_estimates,
        frem_model_results.individual_estimates_covariance,
        parameters=etas,
        seed=rng,
    ).droplevel('sample')
    ninds = len(pool.index.unique())
    ishr = calculate_individual_shrinkage(
        frem_model,
        frem_model_results.parameter_estimates,
        frem_model_results.individual_estimates_covariance,
    )
    ishr = ishr[pool.columns]
    lower_indices = np.tril_indices(len(etas))
    pop_params = np.array(dist.variance).astype(str)[lower_indices]
    parameter_samples = np.empty((samples, len(pop_params)))
    remaining_samples = samples
    k = 0
    while k < remaining_samples:
        bootstrap = pool.sample(n=ninds, replace=True, random_state=rng.bit_generator)
        ishk = ishr.loc[bootstrap.index]
        mean = ishk.to_numpy().mean(0)
        cf = (1 / (1 - mean)) ** (1 / 2)
        corrected_bootstrap = np.multiply(bootstrap.to_numpy(), cf)
        bootstrap_cov = np.cov(np.transpose(corrected_bootstrap))
        if not is_posdef(bootstrap_cov):
            continue
        parameter_samples[k, :] = bootstrap_cov[lower_indices]
        k += 1
    frame = pd.DataFrame(parameter_samples, columns=pop_params)
    # Shift to the mean of the parameter estimate
    shift = frem_model_results.parameter_estimates[pop_params] - frame.mean()
    frame = frame + shift
    res = calculate_results_from_samples(
        frem_model, frem_model_results, continuous, categorical, frame, rescale=rescale
    )
    return res


def psn_reorder_base_model_inits(model, path):
    """Reorder omega inits from base model in PsN

    If base model was reordered PsN writes the omega inits dict to m1/model_1.inits
    """
    order_path = path / 'm1' / 'model_1.inits'
    if order_path.is_file():
        with open(order_path, 'r') as fh:
            lines = fh.readlines()
        lines = lines[1:-1]
        replacements = {}
        for line in lines:
            stripped = line.strip().replace(' ', '').replace("'", '').rstrip(',').replace('>', '')
            a = stripped.split('=')
            coords = a[0][6:-1]
            t = tuple(coords.split(','))
            t = (int(t[0]), int(t[1]))
            try:
                replacements[t] = float(a[1])
            except ValueError:
                pass

        def sortfunc(x):
            return float(x[0]) + float(x[1]) / 1000

        order = sorted(replacements, key=sortfunc)
        values = [replacements[i] for i in order]
        i = 0
        d = {}
        for p in model.parameters:
            if i == len(values):
                break
            if p.name in model.random_variables.parameter_names:
                d[p.name] = values[i]
                i += 1
        newparams = model.parameters.set_initial_estimates(d)
        model = model.replace(parameters=newparams)
    return model


def psn_frem_results(path, force_posdef_covmatrix=False, force_posdef_samples=500, method=None):
    """Create frem results from a PsN FREM run

    :param path: Path to PsN frem run directory
    :return: A :class:`FREMResults` object

    """
    path = Path(path)

    model_4_path = path / 'final_models' / 'model_4.mod'
    if not model_4_path.is_file():
        raise IOError(f'Could not find FREM model 4: {str(model_4_path)}')
    model_4 = Model.parse_model(model_4_path)
    model_4_results = read_modelfit_results(model_4_path)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Adjusting initial estimates")
        if model_4_results is None:
            raise ValueError('Model 4 has no results')
    cov_model = None
    if method == 'cov_sampling':
        try:
            model_4_results.covariance_matrix
        except Exception:
            model_4b_path = path / 'final_models' / 'model_4b.mod'
            try:
                model_4b = Model.parse_model(model_4b_path)
            except FileNotFoundError:
                pass
            else:
                cov_model = model_4b

    with open(path / 'covariates_summary.csv') as covsum:
        covsum.readline()
        raw_cov_list = covsum.readline()
    all_covariates = raw_cov_list[1:].rstrip().split(',')

    # FIXME: Not introducing yaml parser in pharmpy just yet. Options should be collected
    # differently. Perhaps using json
    logtransformed_covariates = []
    rescale = True
    with open(path / 'meta.yaml') as meta:
        for row in meta:
            row = row.strip()
            if row.startswith('rescale: 1'):
                rescale = True
            elif row.startswith('rescale: 0'):
                rescale = False
            if row.startswith("log: ''"):
                logtransformed_covariates = []
            elif row.startswith('log: '):
                logtransformed_covariates = row[5:].split(',')

    # add log transformed columns for the -log option. Should be done when creating dataset
    df = model_4.dataset
    if logtransformed_covariates:
        for lncov in logtransformed_covariates:
            df = df.copy()
            df[f'LN{lncov}'] = np.log(df[lncov])
        model_4 = model_4.replace(dataset=df)

    nunique = get_baselines(model_4)[all_covariates].nunique()
    continuous = list(nunique.index[nunique != 2])
    categorical = list(nunique.index[nunique == 2])

    intmod_names = ['model_1.mod', 'model_2.mod', 'model_3.mod', 'model_3b.mod']
    intmods = []
    intmodres = []
    for m in intmod_names:
        intmod_path = path / 'm1' / m
        if intmod_path.is_file():
            intmod = Model.parse_model(intmod_path)
            intmods.append(intmod)
            res = read_modelfit_results(intmod_path)
            intmodres.append(res)

    model1b = Model.parse_model(path / 'm1' / 'model_1b.mod')
    model1 = intmods[0]
    model1_res = replace(
        intmodres[0], parameter_estimates=pd.Series(model1b.parameters.nonfixed.inits)
    )
    model1 = psn_reorder_base_model_inits(model1, path)
    intmods[0] = model1
    intmodres[0] = model1_res

    res = calculate_results(
        model_4,
        model_4_results,
        continuous,
        categorical,
        method=method,
        force_posdef_covmatrix=force_posdef_covmatrix,
        force_posdef_samples=force_posdef_samples,
        cov_model=cov_model,
        rescale=rescale,
        intermediate_models=intmods,
        intermediate_models_res=intmodres,
        seed=np.random.default_rng(9843),
    )
    return res



import json
import os
import time
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from contextlib import contextmanager
from itertools import chain, groupby
from multiprocessing import get_context
from pathlib import Path
from queue import Queue
from random import random
from threading import Barrier

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.internals.fs.lock import (
    AcquiringProcessLevelLockWouldBlockError,
    AcquiringThreadLevelLockWouldBlockError,
    RecursiveDeadlockError,
    path_lock,
    process_level_path_lock,
    thread_level_lock,
)

mp = get_context(method='spawn')


class SimpleThreadManager:
    @staticmethod
    def Barrier(*args, **kwargs):
        return Barrier(*args, **kwargs)

    @staticmethod
    def Queue(*args, **kwargs):
        return Queue(*args, **kwargs)


@contextmanager
def processes(n: int):
    with ProcessPoolExecutor(max_workers=n, mp_context=mp) as executor:
        with mp.Manager() as m:
            yield executor, m


@contextmanager
def threads(n: int):
    with ThreadPoolExecutor(max_workers=n) as executor:
        yield executor, SimpleThreadManager


@contextmanager
def lock(directory):
    with chdir(directory):
        lock = Path('lock')
        lock.touch()
        path = str(lock)
        yield path


def lock_first(is_locked, is_done, path, shared):
    with path_lock(path, shared=shared, blocking=False):
        is_locked.wait()
        is_done.wait()


def lock_rest(is_locked, is_done, path, shared, blocking):
    is_locked.wait()
    with path_lock(path, shared=shared, blocking=blocking):
        if is_done is not None and shared:
            is_done.wait()


@pytest.mark.parametrize(
    'parallelization, exception',
    (
        (threads, AcquiringThreadLevelLockWouldBlockError),
        (processes, AcquiringProcessLevelLockWouldBlockError),
    ),
)
@pytest.mark.parametrize(
    'shared',
    (
        [False, True],
        [False, False],
        [False, True, False],
        [False, True, True],
        [False, False, True],
        [False, False, False],
        [True, False],
        [True, True, False],
        [True, False, False],
    ),
    ids=repr,
)
def test_non_blocking(tmp_path, parallelization, exception, shared):
    if os.name == 'nt' and 'processes' in repr(parallelization):
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    assert len(shared) >= 2
    assert not shared[0] or not shared[-1]

    with lock(tmp_path) as path:
        n = len(shared)
        nb = 1 + (
            1
            if not shared[0] or (os.name == 'nt' and 'processes' in repr(parallelization))
            else sum(map(lambda s: 1 if s else 0, shared))
        )

        with parallelization(n) as [executor, m]:
            is_locked = m.Barrier(n)
            is_done = m.Barrier(nb)

            tasks = []

            tasks.append(executor.submit(lock_first, is_locked, is_done, path, shared[0]))
            for s in shared[1:-1]:
                tasks.append(
                    executor.submit(
                        lock_rest, is_locked, is_done if nb >= 3 else None, path, s, True
                    )
                )
            tasks.append(executor.submit(lock_rest, is_locked, None, path, shared[-1], False))

            with pytest.raises(exception):
                tasks[-1].result()

            is_done.wait()

            for task in tasks[:-1]:
                task.result()


def locked_threads(parameters, is_done):
    n = len(parameters)
    with threads(n) as (executor, _):
        tasks = []
        for params in parameters:
            tasks.append(executor.submit(*params))
        for task in as_completed(tasks):
            try:
                task.result()
            except:  # noqa E722
                is_done.wait()
                raise


@pytest.mark.parametrize('n_blocking', (0, 1, 2, 3, 4, 5))
def test_non_blocking_processes_and_threads(tmp_path, n_blocking):
    if os.name == 'nt':
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        with processes(2) as (executor, m):
            is_locked = m.Barrier(2 + n_blocking)
            is_done = m.Barrier(2)

            t1 = executor.submit(lock_first, is_locked, is_done, path, False)

            t2 = executor.submit(
                locked_threads,
                ([(lock_rest, is_locked, None, path, True, True)] * n_blocking)
                + [(lock_rest, is_locked, None, path, True, False)],
                is_done,
            )

            with pytest.raises(AcquiringProcessLevelLockWouldBlockError):
                t2.result()

            t1.result()


def lock_shared(are_locked, q, path, i):
    with path_lock(path, shared=True):
        are_locked.wait()
        time.sleep(1)
        q.put(i)


def lock_exclusive(are_locked, q, path, i):
    are_locked.wait()
    with path_lock(path, shared=False):
        q.put(i)


@pytest.mark.parametrize('parallelization', (threads, processes))
def test_many_shared_one_exclusive_blocking(tmp_path, parallelization):
    if os.name == 'nt' and 'processes' in repr(parallelization):
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        n = 10

        with parallelization(n) as [executor, m]:
            results_queue = m.Queue()

            # NOTE: We run the test twice to reuse workers to catch errors where
            # some workers are left in a locked state.
            for _ in range(2):
                are_locked = m.Barrier(n)

                for i in range(1, n):
                    executor.submit(lock_shared, are_locked, results_queue, path, i)
                last = executor.submit(lock_exclusive, are_locked, results_queue, path, 0)

                last.result()

                results = []
                for i in range(n):
                    results.append(results_queue.get())

                assert sorted(results) == sorted(range(n))
                assert results[-1] == 0


@pytest.mark.parametrize('acquire_lock', (path_lock, thread_level_lock, process_level_path_lock))
@pytest.mark.parametrize('shared', (True, False))
def test_non_reentrant_dead_lock(tmp_path, acquire_lock, shared):
    with lock(tmp_path) as path:
        with acquire_lock(path, shared=shared):
            with pytest.raises(RecursiveDeadlockError):
                with acquire_lock(path, shared=shared):
                    pass


@pytest.mark.parametrize('shared', (True, False))
def test_reentrant(tmp_path, shared):
    with lock(tmp_path) as path:
        with path_lock(path, shared=shared, reentrant=False):
            with path_lock(path, shared=shared, reentrant=True):
                pass

        with path_lock(path, shared=shared, reentrant=True):
            with path_lock(path, shared=shared, reentrant=True):
                pass

        with path_lock(path, shared=shared, reentrant=False):
            with path_lock(path, shared=shared, reentrant=True):
                with path_lock(path, shared=shared, reentrant=True):
                    pass


@pytest.mark.parametrize(
    'shared',
    (
        [True, False],
        [False, True],
        [True, False, True],
        [False, True, False],
        [True, True, False],
        [False, False, True],
    ),
    ids=repr,
)
def test_reentrant_mixed(tmp_path, shared):
    with lock(tmp_path) as path:

        def rec(types):
            if types:
                shared, *rest = types
                with path_lock(path, shared=shared, reentrant=True):
                    rec(rest)

        rec(shared)


def exclusive_thread_write(path, i):
    with path_lock(path, shared=False) as fd:
        try:
            with open(fd, closefd=False) as fp:
                fp.seek(0)
                done = json.load(fp)
        except json.decoder.JSONDecodeError:
            done = []

        done.append(i)
        with open(fd, 'w', closefd=False) as fp:
            fp.seek(0)
            fp.truncate(0)
            json.dump(done, fp)
            fp.seek(0)


def many_exclusive_threads_and_processes_rw_process(path, indices):
    with ThreadPoolExecutor(max_workers=len(indices)) as executor:
        executor.map(exclusive_thread_write, [path] * len(indices), indices)


def test_many_exclusive_threads_and_processes_rw(tmp_path):
    if os.name == 'nt':
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        m = 10
        n = m**2
        items = list(range(n))
        partition = list(
            map(
                lambda g: list(map(lambda t: t[1], g[1])),
                groupby(enumerate(items), lambda t: t[0] // m),
            )
        )

        assert len(partition) == m
        assert sorted(chain(*partition)) == items

        with mp.Pool(processes=len(partition)) as pool:
            pool.starmap(
                many_exclusive_threads_and_processes_rw_process,
                map(lambda part: (path, part), partition),
            )

        with path_lock(path, shared=False) as fd:
            with open(fd, closefd=False) as fp:
                fp.seek(0)
                results = json.load(fp)
                fp.seek(0)

        assert sorted(results) == sorted(range(n))


@pytest.mark.parametrize('parallelization, n', ((threads, 200), (processes, 20)))
def test_many_exclusive(tmp_path, parallelization, n):
    if os.name == 'nt' and 'processes' in repr(parallelization):
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        items = list(range(n))

        with parallelization(n) as [executor, _]:
            executor.map(
                exclusive_thread_write,
                [path] * n,
                items,
            )

        with path_lock(path, shared=False) as fd:
            with open(fd, closefd=False) as fp:
                fp.seek(0)
                results = json.load(fp)
                fp.seek(0)

        assert sorted(results) == sorted(range(n))


def lock_shared_chained(path, first_is_locked, recvp, send, recvn, results, n, i):
    if i > 0:
        j = recvp.get()  # Wait for previous thread to be locked
        assert j == i - 1
    with path_lock(path, shared=True):
        if i == 0:
            first_is_locked.wait()  # Allow exclusive lock attempt
        results.put(i)
        send.put(i)  # Notify next thread that we are locked
        send.put(i)  # Notify previous thread that we are locked
        if i < n - 2:
            j = recvn.get()  # Wait for next thread to be locked
            assert j == i + 1


@pytest.mark.parametrize('parallelization', (threads, processes))
def test_chained_shared_one_exclusive_blocking(tmp_path, parallelization):
    if os.name == 'nt' and 'processes' in repr(parallelization):
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        n = 10

        with parallelization(n) as [executor, m]:
            results_queue = m.Queue()

            # NOTE: We run the test twice to reuse workers to catch errors where
            # some workers are left in a locked state.
            for _ in range(2):
                first_is_locked = m.Barrier(2)
                queues = [m.Queue() for _ in range(n + 1)]

                for i in range(n - 1):
                    executor.submit(
                        lock_shared_chained,
                        path,
                        first_is_locked,
                        queues[i],
                        queues[i + 1],
                        queues[i + 2],
                        results_queue,
                        n,
                        i,
                    )
                last = executor.submit(lock_exclusive, first_is_locked, results_queue, path, n - 1)

                last.result()

                results = []
                for i in range(n):
                    results.append(results_queue.get())

                assert results == sorted(range(n))


def sync_read(path, filename, before, after, q, i):
    before.wait()
    time.sleep(0.01 * random())
    with path_lock(path, shared=True):
        after.wait()
        with open(filename) as fp:
            contents = json.load(fp)
    q.put({'type': 'read', 'id': i, 'contents': contents})


def sync_write(path, filename, q, i):
    with path_lock(path, shared=False):
        with open(filename) as fp:
            contents = json.load(fp)

        contents['id'] = i
        contents['counter'] += 1

        with open(filename, 'w') as fp:
            json.dump(contents, fp)

        with open(filename) as fp:
            contents = json.load(fp)

        contents['copy'] += 1

        with open(filename, 'w') as fp:
            json.dump(contents, fp)

    q.put({'type': 'write', 'id': i, 'contents': contents})


@pytest.mark.parametrize('parallelization', (threads, processes))
def test_synchronized_reads_blocking(tmp_path, parallelization):
    if os.name == 'nt' and 'processes' in repr(parallelization):
        pytest.skip("TODO Processes-based tests randomly fail on Windows.")

    with lock(tmp_path) as path:
        filename = 'rw'

        with open(filename, 'w') as fp:
            json.dump({'id': -1, 'counter': 0, 'copy': 0}, fp)

        n = 1000
        k = 7
        p = 10
        max_write = p - k

        with parallelization(p) as [executor, m]:
            q = m.Queue()
            i = 0
            g = 0
            w = 0
            messages = []
            group = {}
            while i < n:
                c = min(n - i, k)
                if w >= max_write or random() < 1 / (c + 1):
                    before = m.Barrier(c)
                    after = m.Barrier(c)
                    for _ in range(c):
                        executor.submit(sync_read, path, filename, before, after, q, i)
                        group[i] = g
                        i += 1
                    g += 1
                else:
                    executor.submit(sync_write, path, filename, q, i)
                    i += 1
                    w += 1

                while (
                    not q.empty() or len(messages) < i - k
                ):  # NOTE: We empty the queue as much as possible
                    messages.append(q.get())
                    if messages[-1]['type'] == 'write':
                        w -= 1

            while len(messages) < n:
                messages.append(q.get())

        values = {}
        for message in messages:
            t = message['type']
            if t == 'read':
                i = message['id']
                g = group[i]
                counter = message['contents']['counter']
                # NOTE: counter is the same for the whole group
                assert values.setdefault(g, counter) == counter
                # NOTE: copy is identical to counter
                assert message['contents']['copy'] == counter
            else:
                assert t == 'write'
                # NOTE: check nobody else wrote to file while we held the lock
                assert message['id'] == message['contents']['id']
                assert message['contents']['copy'] == message['contents']['counter']



from typing import Callable, Iterable, TypeVar

from .reachability import reachable_from

T = TypeVar('T')


def strongly_connected_component_of(
    vertex: T, successors: Callable[[T], Iterable[T]], predecessors: Callable[[T], Iterable[T]]
):
    forward_reachable = reachable_from({vertex}, successors)

    # NOTE: This searches for backward reachable vertices on the graph induced
    # by the forward reachable vertices and is equivalent to (but less wasteful
    # than) first computing the backward reachable vertices on the original
    # graph and then computing the intersection with the forward reachable
    # vertices.
    return reachable_from(
        {vertex},
        lambda u: filter(forward_reachable.__contains__, predecessors(u)),
    )



from typing import TYPE_CHECKING

from pharmpy.internals.math import cov2corr

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd


def calculate_se_from_cov(cov: pd.DataFrame):
    """Calculate standard errors from a covariance matrix

    Parameters
    ----------
    cov : pd.DataFrame
        Input covariance matrix

    Return
    ------
    pd.Series
        Standard errors

    Examples
    --------
    >>> from pharmpy.modeling import calculate_se_from_cov
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> cov = results.covariance_matrix
    >>> cov
                      PTVCL          PTVV   THETA_3      IVCL           IVV     SIGMA_1_1
    PTVCL      4.411510e-08  4.010000e-08 -0.000002 -0.000001  1.538630e-07  8.178090e-08
    PTVV       4.010000e-08  7.233530e-04 -0.000804  0.000050  7.171840e-05  1.461760e-05
    THETA_3   -1.665010e-06 -8.040250e-04  0.007016 -0.000108 -3.944800e-05  2.932950e-05
    IVCL      -1.093430e-06  4.981380e-05 -0.000108  0.000180 -1.856650e-05  4.867230e-06
    IVV        1.538630e-07  7.171840e-05 -0.000039 -0.000019  5.589820e-05 -4.685650e-07
    SIGMA_1_1  8.178090e-08  1.461760e-05  0.000029  0.000005 -4.685650e-07  5.195640e-06
    >>> calculate_se_from_cov(cov)
    PTVCL        0.000210
    PTVV         0.026895
    THETA_3      0.083762
    IVCL         0.013415
    IVV          0.007477
    SIGMA_1_1    0.002279
    dtype: float64

    See also
    --------
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """
    se = pd.Series(np.sqrt(np.diag(cov.values)), index=cov.index)
    return se


def calculate_se_from_prec(precision_matrix: pd.DataFrame):
    """Calculate standard errors from a precision matrix

    Parameters
    ----------
    precision_matrix : pd.DataFrame
        Input precision matrix

    Return
    ------
    pd.Series
        Standard errors

    Examples
    --------
    >>> from pharmpy.modeling import calculate_se_from_prec
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> prec = results.precision_matrix
    >>> prec
                      PTVCL          PTVV       THETA_3           IVCL           IVV      SIGMA_1_1
    PTVCL      2.995567e+07  22660.028196  16057.855248  203511.614428 -39474.250514 -820118.299536
    PTVV       2.266003e+04   2129.904642    260.176234    -375.266233  -2800.816246   -7718.769557
    THETA_3    1.605786e+04    260.176234    187.038903     177.207683   -205.808480   -2225.150449
    IVCL       2.035116e+05   -375.266233    177.207683    7527.530027   2462.974821   -9977.488860
    IVV       -3.947425e+04  -2800.816246   -205.808480    2462.974821  22343.198618    9370.758371
    SIGMA_1_1 -8.201183e+05  -7718.769557  -2225.150449   -9977.488860   9370.758371  249847.177845
    >>> calculate_se_from_prec(prec)
    PTVCL        0.000210
    PTVV         0.026895
    THETA_3      0.083762
    IVCL         0.013415
    IVV          0.007477
    SIGMA_1_1    0.002279
    dtype: float64

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """
    se = pd.Series(
        np.sqrt(np.diag(np.linalg.inv(precision_matrix.values))), index=precision_matrix.index
    )
    return se


def calculate_corr_from_cov(cov: pd.DataFrame):
    """Calculate correlation matrix from a covariance matrix

    Parameters
    ----------
    cov : pd.DataFrame
        Covariance matrix

    Return
    ------
    pd.DataFrame
        Correlation matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_corr_from_cov
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> cov = results.covariance_matrix
    >>> cov
                      PTVCL          PTVV   THETA_3      IVCL           IVV     SIGMA_1_1
    PTVCL      4.411510e-08  4.010000e-08 -0.000002 -0.000001  1.538630e-07  8.178090e-08
    PTVV       4.010000e-08  7.233530e-04 -0.000804  0.000050  7.171840e-05  1.461760e-05
    THETA_3   -1.665010e-06 -8.040250e-04  0.007016 -0.000108 -3.944800e-05  2.932950e-05
    IVCL      -1.093430e-06  4.981380e-05 -0.000108  0.000180 -1.856650e-05  4.867230e-06
    IVV        1.538630e-07  7.171840e-05 -0.000039 -0.000019  5.589820e-05 -4.685650e-07
    SIGMA_1_1  8.178090e-08  1.461760e-05  0.000029  0.000005 -4.685650e-07  5.195640e-06
    >>> calculate_corr_from_cov(cov)
                  PTVCL      PTVV   THETA_3      IVCL       IVV  SIGMA_1_1
    PTVCL      1.000000  0.007099 -0.094640 -0.388059  0.097981   0.170820
    PTVV       0.007099  1.000000 -0.356899  0.138062  0.356662   0.238441
    THETA_3   -0.094640 -0.356899  1.000000 -0.096515 -0.062991   0.153616
    IVCL      -0.388059  0.138062 -0.096515  1.000000 -0.185111   0.159170
    IVV        0.097981  0.356662 -0.062991 -0.185111  1.000000  -0.027495
    SIGMA_1_1  0.170820  0.238441  0.153616  0.159170 -0.027495   1.000000

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """

    corr = pd.DataFrame(cov2corr(cov.values), index=cov.index, columns=cov.columns)
    return corr


def calculate_cov_from_prec(precision_matrix: pd.DataFrame):
    """Calculate covariance matrix from a precision matrix

    Parameters
    ----------
    precision_matrix : pd.DataFrame
        Precision matrix

    Return
    ------
    pd.DataFrame
        Covariance matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_cov_from_prec
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> prec = results.precision_matrix
    >>> prec
                      PTVCL          PTVV       THETA_3           IVCL           IVV      SIGMA_1_1
    PTVCL      2.995567e+07  22660.028196  16057.855248  203511.614428 -39474.250514 -820118.299536
    PTVV       2.266003e+04   2129.904642    260.176234    -375.266233  -2800.816246   -7718.769557
    THETA_3    1.605786e+04    260.176234    187.038903     177.207683   -205.808480   -2225.150449
    IVCL       2.035116e+05   -375.266233    177.207683    7527.530027   2462.974821   -9977.488860
    IVV       -3.947425e+04  -2800.816246   -205.808480    2462.974821  22343.198618    9370.758371
    SIGMA_1_1 -8.201183e+05  -7718.769557  -2225.150449   -9977.488860   9370.758371  249847.177845
    >>> calculate_cov_from_prec(prec)
                      PTVCL          PTVV   THETA_3      IVCL           IVV     SIGMA_1_1
    PTVCL      4.411510e-08  4.010000e-08 -0.000002 -0.000001  1.538630e-07  8.178090e-08
    PTVV       4.010000e-08  7.233530e-04 -0.000804  0.000050  7.171840e-05  1.461760e-05
    THETA_3   -1.665010e-06 -8.040250e-04  0.007016 -0.000108 -3.944800e-05  2.932950e-05
    IVCL      -1.093430e-06  4.981380e-05 -0.000108  0.000180 -1.856650e-05  4.867230e-06
    IVV        1.538630e-07  7.171840e-05 -0.000039 -0.000019  5.589820e-05 -4.685650e-07
    SIGMA_1_1  8.178090e-08  1.461760e-05  0.000029  0.000005 -4.685650e-07  5.195640e-06

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """

    cov = pd.DataFrame(
        np.linalg.inv(precision_matrix.values),
        index=precision_matrix.index,
        columns=precision_matrix.columns,
    )
    return cov


def calculate_cov_from_corrse(corr: pd.DataFrame, se: pd.Series):
    """Calculate covariance matrix from a correlation matrix and standard errors

    Parameters
    ----------
    corr : pd.DataFrame
        Correlation matrix
    se : pd.Series
        Standard errors

    Return
    ------
    pd.DataFrame
        Covariance matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_cov_from_corrse
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> corr = results.correlation_matrix
    >>> se = results.standard_errors
    >>> corr
                  PTVCL      PTVV   THETA_3      IVCL       IVV  SIGMA_1_1
    PTVCL      1.000000  0.007099 -0.094640 -0.388059  0.097981   0.170820
    PTVV       0.007099  1.000000 -0.356899  0.138062  0.356662   0.238441
    THETA_3   -0.094640 -0.356899  1.000000 -0.096515 -0.062991   0.153616
    IVCL      -0.388059  0.138062 -0.096515  1.000000 -0.185111   0.159170
    IVV        0.097981  0.356662 -0.062991 -0.185111  1.000000  -0.027495
    SIGMA_1_1  0.170820  0.238441  0.153616  0.159170 -0.027495   1.000000
    >>> calculate_cov_from_corrse(corr, se)
                      PTVCL          PTVV   THETA_3      IVCL           IVV     SIGMA_1_1
    PTVCL      4.411512e-08  4.009998e-08 -0.000002 -0.000001  1.538630e-07  8.178111e-08
    PTVV       4.009998e-08  7.233518e-04 -0.000804  0.000050  7.171834e-05  1.461762e-05
    THETA_3   -1.665011e-06 -8.040245e-04  0.007016 -0.000108 -3.944801e-05  2.932957e-05
    IVCL      -1.093431e-06  4.981380e-05 -0.000108  0.000180 -1.856651e-05  4.867245e-06
    IVV        1.538630e-07  7.171834e-05 -0.000039 -0.000019  5.589820e-05 -4.685661e-07
    SIGMA_1_1  8.178111e-08  1.461762e-05  0.000029  0.000005 -4.685661e-07  5.195664e-06

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """

    sd_matrix = np.diag(se.values)
    cov = sd_matrix @ corr.values @ sd_matrix
    cov_df = pd.DataFrame(cov, index=corr.index, columns=corr.columns)
    return cov_df


def calculate_prec_from_cov(cov: pd.DataFrame):
    """Calculate precision matrix from a covariance matrix

    Parameters
    ----------
    cov : pd.DataFrame
        Covariance matrix

    Return
    ------
    pd.DataFrame
        Precision matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_prec_from_cov
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> cov = results.covariance_matrix
    >>> cov
                      PTVCL          PTVV   THETA_3      IVCL           IVV     SIGMA_1_1
    PTVCL      4.411510e-08  4.010000e-08 -0.000002 -0.000001  1.538630e-07  8.178090e-08
    PTVV       4.010000e-08  7.233530e-04 -0.000804  0.000050  7.171840e-05  1.461760e-05
    THETA_3   -1.665010e-06 -8.040250e-04  0.007016 -0.000108 -3.944800e-05  2.932950e-05
    IVCL      -1.093430e-06  4.981380e-05 -0.000108  0.000180 -1.856650e-05  4.867230e-06
    IVV        1.538630e-07  7.171840e-05 -0.000039 -0.000019  5.589820e-05 -4.685650e-07
    SIGMA_1_1  8.178090e-08  1.461760e-05  0.000029  0.000005 -4.685650e-07  5.195640e-06
    >>> calculate_prec_from_cov(cov)
                      PTVCL          PTVV       THETA_3           IVCL           IVV      SIGMA_1_1
    PTVCL      2.995567e+07  22660.028196  16057.855248  203511.614428 -39474.250514 -820118.299536
    PTVV       2.266003e+04   2129.904642    260.176234    -375.266233  -2800.816246   -7718.769557
    THETA_3    1.605786e+04    260.176234    187.038903     177.207683   -205.808480   -2225.150449
    IVCL       2.035116e+05   -375.266233    177.207683    7527.530027   2462.974821   -9977.488860
    IVV       -3.947425e+04  -2800.816246   -205.808480    2462.974821  22343.198618    9370.758371
    SIGMA_1_1 -8.201183e+05  -7718.769557  -2225.150449   -9977.488860   9370.758371  249847.177845

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """

    Pm = pd.DataFrame(np.linalg.inv(cov.values), index=cov.index, columns=cov.columns)
    return Pm


def calculate_prec_from_corrse(corr: pd.DataFrame, se: pd.Series):
    """Calculate precision matrix from a correlation matrix and standard errors

    Parameters
    ----------
    corr : pd.DataFrame
        Correlation matrix
    se : pd.Series
        Standard errors

    Return
    ------
    pd.DataFrame
        Precision matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_prec_from_corrse
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> corr = results.correlation_matrix
    >>> se = results.standard_errors
    >>> corr
                  PTVCL      PTVV   THETA_3      IVCL       IVV  SIGMA_1_1
    PTVCL      1.000000  0.007099 -0.094640 -0.388059  0.097981   0.170820
    PTVV       0.007099  1.000000 -0.356899  0.138062  0.356662   0.238441
    THETA_3   -0.094640 -0.356899  1.000000 -0.096515 -0.062991   0.153616
    IVCL      -0.388059  0.138062 -0.096515  1.000000 -0.185111   0.159170
    IVV        0.097981  0.356662 -0.062991 -0.185111  1.000000  -0.027495
    SIGMA_1_1  0.170820  0.238441  0.153616  0.159170 -0.027495   1.000000
    >>> calculate_prec_from_corrse(corr, se)
                      PTVCL          PTVV       THETA_3           IVCL           IVV      SIGMA_1_1
    PTVCL      2.995565e+07  22660.041788  16057.848052  203511.410335 -39474.240358 -820116.179011
    PTVV       2.266004e+04   2129.908225    260.176399    -375.266263  -2800.818557   -7718.757955
    THETA_3    1.605785e+04    260.176399    187.038825     177.207512   -205.808434   -2225.144772
    IVCL       2.035114e+05   -375.266263    177.207512    7527.518562   2462.972906   -9977.457873
    IVV       -3.947424e+04  -2800.818557   -205.808434    2462.972906  22343.197906    9370.736254
    SIGMA_1_1 -8.201162e+05  -7718.757955  -2225.144772   -9977.457873   9370.736254  249846.006431

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_corr_from_prec : Correlation matrix from precision matrix
    """

    sd_matrix = np.diag(se.values)
    cov = sd_matrix @ corr.values @ sd_matrix
    Pm = pd.DataFrame(np.linalg.inv(cov), index=corr.index, columns=corr.columns)
    return Pm


def calculate_corr_from_prec(precision_matrix: pd.DataFrame):
    """Calculate correlation matrix from a precision matrix

    Parameters
    ----------
    precision_matrix : pd.DataFrame
        Precision matrix

    Return
    ------
    pd.DataFrame
        Correlation matrix

    Examples
    --------
    >>> from pharmpy.modeling import calculate_corr_from_prec
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> prec = results.precision_matrix
    >>> prec
                      PTVCL          PTVV       THETA_3           IVCL           IVV      SIGMA_1_1
    PTVCL      2.995567e+07  22660.028196  16057.855248  203511.614428 -39474.250514 -820118.299536
    PTVV       2.266003e+04   2129.904642    260.176234    -375.266233  -2800.816246   -7718.769557
    THETA_3    1.605786e+04    260.176234    187.038903     177.207683   -205.808480   -2225.150449
    IVCL       2.035116e+05   -375.266233    177.207683    7527.530027   2462.974821   -9977.488860
    IVV       -3.947425e+04  -2800.816246   -205.808480    2462.974821  22343.198618    9370.758371
    SIGMA_1_1 -8.201183e+05  -7718.769557  -2225.150449   -9977.488860   9370.758371  249847.177845
    >>> calculate_corr_from_prec(prec)
                  PTVCL      PTVV   THETA_3      IVCL       IVV  SIGMA_1_1
    PTVCL      1.000000  0.007099 -0.094640 -0.388059  0.097981   0.170820
    PTVV       0.007099  1.000000 -0.356899  0.138062  0.356662   0.238441
    THETA_3   -0.094640 -0.356899  1.000000 -0.096515 -0.062991   0.153616
    IVCL      -0.388059  0.138062 -0.096515  1.000000 -0.185111   0.159170
    IVV        0.097981  0.356662 -0.062991 -0.185111  1.000000  -0.027495
    SIGMA_1_1  0.170820  0.238441  0.153616  0.159170 -0.027495   1.000000

    See also
    --------
    calculate_se_from_cov : Standard errors from covariance matrix
    calculate_se_from_prec : Standard errors from precision matrix
    calculate_corr_from_cov : Correlation matrix from covariance matrix
    calculate_cov_from_prec : Covariance matrix from precision matrix
    calculate_cov_from_corrse : Covariance matrix from correlation matrix and standard errors
    calculate_prec_from_cov : Precision matrix from covariance matrix
    calculate_prec_from_corrse : Precision matrix from correlation matrix and standard errors
    """

    corr = pd.DataFrame(
        cov2corr(np.linalg.inv(precision_matrix.values)),
        index=precision_matrix.index,
        columns=precision_matrix.columns,
    )
    return corr



"""Pharmpy configuration
"""

import configparser
import os
from abc import ABC
from pathlib import Path

import appdirs

appname = 'Pharmpy'
configuration_filename = 'pharmpy.conf'


def user_config_file_enabled():
    return not int(os.getenv('PHARMPYNOCONFIGFILE', 0))


def user_config_path():
    user_path = Path(appdirs.user_config_dir(appname)) / configuration_filename
    return user_path


def site_config_path():
    site_path = Path(appdirs.site_config_dir(appname)) / configuration_filename
    return site_path


def env_config_path():
    env_path = os.getenv('PHARMPYCONFIGPATH')
    if env_path is not None:
        env_path = Path(env_path) / configuration_filename
        if not env_path.is_file():
            raise ValueError(
                'Environment variable PHARMPYCONFIGPATH is set but directory does '
                'not contain a configuration file'
            )
    return env_path


def config_path():
    env_path = env_config_path()
    if env_path is not None:
        return env_path
    else:
        user_path = user_config_path()
        if user_path.is_file():
            return user_path
        else:
            site_path = site_config_path()
            if site_path.is_file():
                return site_path
    return None


def read_configuration():
    config = configparser.ConfigParser()
    path = config_path()
    if path is not None:
        config.read(path)
    return config


config_file = read_configuration()


class ConfigItem:
    def __init__(self, default, description, cls=None):
        self.default = default
        self.__doc__ = description
        if cls is None:
            self.cls = type(self.default)
        else:
            self.cls = cls

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance, value):
        try:
            if isinstance(self.default, list) and isinstance(value, str):
                value = [v.strip() for v in value.split(',')]
            if isinstance(self.default, bool) and isinstance(value, str):
                if value.lower() == 'true':
                    value = True
                else:
                    value = False
            instance.__dict__[self.name] = self.cls(value)
        except ValueError as exc:
            raise TypeError(
                f'Trying to set configuration item {self.name} using object of wrong '
                f'type: {type(value)} is not {self.cls}'
            ) from exc


class Configuration(ABC):
    module: str

    def __init__(self):
        if user_config_file_enabled() and self.module in config_file.keys():
            for key, value in config_file[self.module].items():
                setattr(self, key, value)

    def __str__(self):
        settings = ''
        for key, value in vars(self).items():
            settings += f"{key}:\t{value}\n"
        return settings


class ConfigurationContext:
    """Context to temporarily set configuration options"""

    def __init__(self, config, **kwargs):
        self.config = config
        self.options = kwargs

    def __enter__(self):
        old = {}
        for key in self.options.keys():
            old[key] = getattr(self.config, key)
        for key, val in self.options.items():
            setattr(self.config, key, val)
        self.old = old
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for key, val in self.old.items():
            setattr(self.config, key, val)



from .allometry import add_allometry
from .basic_models import create_basic_pk_model
from .blq import transform_blq
from .common import (
    bump_model_number,
    convert_model,
    create_config_template,
    filter_dataset,
    get_config_path,
    get_model_code,
    get_model_covariates,
    load_example_model,
    print_model_code,
    print_model_symbols,
    read_model,
    read_model_from_string,
    remove_unused_parameters_and_rvs,
    rename_symbols,
    set_name,
    write_model,
)
from .compartments import get_bioavailability, get_lag_times
from .covariate_effect import (
    add_covariate_effect,
    get_covariates,
    has_covariate_effect,
    remove_covariate_effect,
)
from .data import (
    add_admid,
    add_cmt,
    add_time_after_dose,
    bin_observations,
    check_dataset,
    deidentify_data,
    drop_columns,
    drop_dropped_columns,
    expand_additional_doses,
    get_admid,
    get_baselines,
    get_cmt,
    get_concentration_parameters_from_data,
    get_covariate_baselines,
    get_doseid,
    get_doses,
    get_evid,
    get_ids,
    get_mdv,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
    list_time_varying_covariates,
    load_dataset,
    read_dataset_from_datainfo,
    remove_loq_data,
    set_covariates,
    set_dvid,
    set_lloq_data,
    set_reference_values,
    translate_nmtran_time,
    undrop_columns,
    unload_dataset,
)
from .error import (
    has_additive_error_model,
    has_combined_error_model,
    has_proportional_error_model,
    has_weighted_error_model,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_dtbs_error_model,
    set_iiv_on_ruv,
    set_power_on_ruv,
    set_proportional_error_model,
    set_time_varying_error_model,
    set_weighted_error_model,
    use_thetas_for_error_stdev,
)
from .estimation import calculate_parameters_from_ucp, calculate_ucp_scale
from .estimation_steps import (
    add_estimation_step,
    add_parameter_uncertainty_step,
    append_estimation_step_options,
    remove_estimation_step,
    remove_parameter_uncertainty_step,
    set_estimation_step,
    set_evaluation_step,
    set_simulation,
)
from .evaluation import (
    evaluate_epsilon_gradient,
    evaluate_eta_gradient,
    evaluate_expression,
    evaluate_individual_prediction,
    evaluate_population_prediction,
    evaluate_weighted_residuals,
)
from .expressions import (
    calculate_epsilon_gradient_expression,
    calculate_eta_gradient_expression,
    cleanup_model,
    create_symbol,
    get_dv_symbol,
    get_individual_parameters,
    get_individual_prediction_expression,
    get_observation_expression,
    get_parameter_rv,
    get_pd_parameters,
    get_pk_parameters,
    get_population_prediction_expression,
    get_rv_parameters,
    greekify_model,
    has_random_effect,
    is_linearized,
    is_real,
    make_declarative,
    mu_reference_model,
    simplify_expression,
)
from .iterators import omit_data, resample_data
from .math import (
    calculate_corr_from_cov,
    calculate_corr_from_prec,
    calculate_cov_from_corrse,
    calculate_cov_from_prec,
    calculate_prec_from_corrse,
    calculate_prec_from_cov,
    calculate_se_from_cov,
    calculate_se_from_prec,
)
from .metabolite import add_metabolite, has_presystemic_metabolite
from .odes import (
    add_bioavailability,
    add_individual_parameter,
    add_lag_time,
    add_peripheral_compartment,
    display_odes,
    find_clearance_parameters,
    find_volume_parameters,
    get_central_volume_and_clearance,
    get_initial_conditions,
    get_number_of_peripheral_compartments,
    get_number_of_transit_compartments,
    get_zero_order_inputs,
    has_first_order_absorption,
    has_first_order_elimination,
    has_instantaneous_absorption,
    has_linear_odes,
    has_linear_odes_with_real_eigenvalues,
    has_michaelis_menten_elimination,
    has_mixed_mm_fo_elimination,
    has_odes,
    has_seq_zo_fo_absorption,
    has_zero_order_absorption,
    has_zero_order_elimination,
    remove_bioavailability,
    remove_lag_time,
    remove_peripheral_compartment,
    set_first_order_absorption,
    set_first_order_elimination,
    set_initial_condition,
    set_instantaneous_absorption,
    set_michaelis_menten_elimination,
    set_mixed_mm_fo_elimination,
    set_ode_solver,
    set_peripheral_compartments,
    set_seq_zo_fo_absorption,
    set_transit_compartments,
    set_zero_order_absorption,
    set_zero_order_elimination,
    set_zero_order_input,
    solve_ode_system,
)
from .parameter_sampling import (
    create_rng,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    sample_parameters_uniformly,
)
from .parameter_variability import (
    add_iiv,
    add_iov,
    add_pd_iiv,
    add_pk_iiv,
    create_joint_distribution,
    remove_iiv,
    remove_iov,
    split_joint_distribution,
    transform_etas_boxcox,
    transform_etas_john_draper,
    transform_etas_tdist,
    update_initial_individual_estimates,
)
from .parameters import (
    add_population_parameter,
    fix_or_unfix_parameters,
    fix_parameters,
    fix_parameters_to,
    get_omegas,
    get_sigmas,
    get_thetas,
    set_initial_estimates,
    set_lower_bounds,
    set_upper_bounds,
    unconstrain_parameters,
    unfix_parameters,
    unfix_parameters_to,
    update_inits,
)
from .pd import add_effect_compartment, add_indirect_effect, set_baseline_effect, set_direct_effect
from .plots import (
    plot_abs_cwres_vs_ipred,
    plot_cwres_vs_idv,
    plot_dv_vs_ipred,
    plot_dv_vs_pred,
    plot_individual_predictions,
    plot_iofv_vs_iofv,
    plot_transformed_eta_distributions,
)
from .random_variables import replace_non_random_rvs
from .results import (
    calculate_aic,
    calculate_bic,
    calculate_eta_shrinkage,
    calculate_individual_parameter_statistics,
    calculate_individual_shrinkage,
    calculate_pk_parameters_statistics,
    check_high_correlations,
    check_parameters_near_bounds,
)
from .tmdd import set_tmdd
from .units import get_unit_of
from .write_csv import write_csv

# Must be set directly, otherwise errors about unused imports
__all__ = [
    'add_admid',
    'add_cmt',
    'add_allometry',
    'add_parameter_uncertainty_step',
    'add_covariate_effect',
    'add_estimation_step',
    'add_effect_compartment',
    'add_indirect_effect',
    'add_iiv',
    'add_individual_parameter',
    'add_iov',
    'add_lag_time',
    'add_bioavailability',
    'add_metabolite',
    'add_peripheral_compartment',
    'add_pk_iiv',
    'add_pd_iiv',
    'add_population_parameter',
    'add_time_after_dose',
    'append_estimation_step_options',
    'bin_observations',
    'bump_model_number',
    'calculate_aic',
    'calculate_bic',
    'calculate_corr_from_cov',
    'calculate_corr_from_prec',
    'calculate_cov_from_corrse',
    'calculate_cov_from_prec',
    'calculate_epsilon_gradient_expression',
    'calculate_eta_gradient_expression',
    'calculate_eta_shrinkage',
    'calculate_individual_parameter_statistics',
    'calculate_individual_shrinkage',
    'calculate_prec_from_corrse',
    'calculate_prec_from_cov',
    'calculate_parameters_from_ucp',
    'calculate_pk_parameters_statistics',
    'calculate_se_from_cov',
    'calculate_se_from_prec',
    'calculate_ucp_scale',
    'check_dataset',
    'check_high_correlations',
    'check_parameters_near_bounds',
    'cleanup_model',
    'convert_model',
    'create_basic_pk_model',
    'create_config_template',
    'create_joint_distribution',
    'create_rng',
    'create_symbol',
    'deidentify_data',
    'display_odes',
    'drop_columns',
    'drop_dropped_columns',
    'evaluate_epsilon_gradient',
    'evaluate_eta_gradient',
    'evaluate_expression',
    'evaluate_individual_prediction',
    'evaluate_population_prediction',
    'evaluate_weighted_residuals',
    'expand_additional_doses',
    'filter_dataset',
    'find_clearance_parameters',
    'find_volume_parameters',
    'fix_or_unfix_parameters',
    'fix_parameters',
    'fix_parameters_to',
    'get_admid',
    'get_baselines',
    'get_bioavailability',
    'get_central_volume_and_clearance',
    'get_cmt',
    'get_concentration_parameters_from_data',
    'get_config_path',
    'get_covariates',
    'get_covariate_baselines',
    'get_doses',
    'get_doseid',
    'get_dv_symbol',
    'get_evid',
    'get_ids',
    'get_individual_parameters',
    'get_individual_prediction_expression',
    'get_initial_conditions',
    'get_lag_times',
    'get_mdv',
    'get_model_code',
    'get_model_covariates',
    'get_number_of_individuals',
    'get_number_of_observations',
    'get_number_of_observations_per_individual',
    'get_omegas',
    'get_observations',
    'get_observation_expression',
    'get_pk_parameters',
    'get_pd_parameters',
    'get_population_prediction_expression',
    'get_rv_parameters',
    'get_parameter_rv',
    'get_sigmas',
    'get_thetas',
    'get_unit_of',
    'get_zero_order_inputs',
    'greekify_model',
    'has_additive_error_model',
    'has_combined_error_model',
    'has_covariate_effect',
    'has_first_order_elimination',
    'has_linear_odes',
    'has_linear_odes_with_real_eigenvalues',
    'has_michaelis_menten_elimination',
    'has_mixed_mm_fo_elimination',
    'has_odes',
    'has_presystemic_metabolite',
    'has_proportional_error_model',
    'has_random_effect',
    'has_zero_order_absorption',
    'has_first_order_absorption',
    'has_instantaneous_absorption',
    'has_seq_zo_fo_absorption',
    'get_number_of_peripheral_compartments',
    'get_number_of_transit_compartments',
    'has_zero_order_elimination',
    'has_weighted_error_model',
    'is_linearized',
    'is_real',
    'list_time_varying_covariates',
    'load_dataset',
    'load_example_model',
    'make_declarative',
    'mu_reference_model',
    'omit_data',
    'plot_abs_cwres_vs_ipred',
    'plot_cwres_vs_idv',
    'plot_dv_vs_ipred',
    'plot_dv_vs_pred',
    'plot_individual_predictions',
    'plot_iofv_vs_iofv',
    'plot_transformed_eta_distributions',
    'print_model_code',
    'print_model_symbols',
    'read_dataset_from_datainfo',
    'read_model',
    'read_model_from_string',
    'rename_symbols',
    'remove_parameter_uncertainty_step',
    'remove_covariate_effect',
    'remove_error_model',
    'remove_estimation_step',
    'remove_iiv',
    'remove_iov',
    'remove_bioavailability',
    'remove_lag_time',
    'remove_loq_data',
    'remove_peripheral_compartment',
    'remove_unused_parameters_and_rvs',
    'replace_non_random_rvs',
    'resample_data',
    'sample_parameters_from_covariance_matrix',
    'sample_individual_estimates',
    'sample_parameters_uniformly',
    'set_additive_error_model',
    'set_baseline_effect',
    'set_instantaneous_absorption',
    'set_combined_error_model',
    'set_covariates',
    'set_direct_effect',
    'set_dtbs_error_model',
    'set_dvid',
    'set_estimation_step',
    'set_evaluation_step',
    'set_simulation',
    'set_first_order_absorption',
    'set_first_order_elimination',
    'set_iiv_on_ruv',
    'set_initial_condition',
    'set_initial_estimates',
    'set_lloq_data',
    'set_lower_bounds',
    'set_upper_bounds',
    'set_michaelis_menten_elimination',
    'set_mixed_mm_fo_elimination',
    'set_name',
    'set_ode_solver',
    'set_peripheral_compartments',
    'set_power_on_ruv',
    'set_proportional_error_model',
    'set_reference_values',
    'set_seq_zo_fo_absorption',
    'set_time_varying_error_model',
    'set_tmdd',
    'set_transit_compartments',
    'set_weighted_error_model',
    'set_zero_order_absorption',
    'set_zero_order_elimination',
    'set_zero_order_input',
    'simplify_expression',
    'solve_ode_system',
    'split_joint_distribution',
    'transform_blq',
    'transform_etas_boxcox',
    'transform_etas_john_draper',
    'transform_etas_tdist',
    'translate_nmtran_time',
    'unfix_parameters',
    'unfix_parameters_to',
    'update_initial_individual_estimates',
    'update_inits',
    'use_thetas_for_error_stdev',
    'write_csv',
    'write_model',
    'unconstrain_parameters',
    'undrop_columns',
    'unload_dataset',
]



"""
:meta private:
"""
from pathlib import Path
from typing import Optional

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fs.path import path_absolute
from pharmpy.model import (
    Assignment,
    ColumnInfo,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    DataInfo,
    EstimationStep,
    EstimationSteps,
    Model,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
    output,
)

from .data import add_admid, read_dataset_from_datainfo
from .error import set_proportional_error_model
from .odes import add_bioavailability, set_first_order_absorption
from .parameter_variability import add_iiv, create_joint_distribution
from .parameters import set_initial_estimates


def create_basic_pk_model(
    administration: str = 'iv',
    dataset_path: Optional[str] = None,
    cl_init: float = 0.01,
    vc_init: float = 1.0,
    mat_init: float = 0.1,
) -> Model:
    """
    Creates a basic pk model of given type

    Parameters
    ----------
    administration : str
        Type of PK model to create. Supported are 'iv', 'oral' and 'ivoral'
    dataset_path : str
        Optional path to a dataset
    cl_init : float
        Initial estimate of the clearance parameter
    vc_init : float
        Initial estimate of the central volume parameter
    mat_init : float
        Initial estimate of the mean absorption time parameter (if applicable)

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = create_basic_pk_model('oral')

    """
    if dataset_path is not None:
        dataset_path = Path(dataset_path)
        di = _create_default_datainfo(dataset_path)
        df = read_dataset_from_datainfo(di, datatype='nonmem')
    else:
        di = DataInfo()
        df = None

    if administration not in [
        'iv',
        'oral',
        'ivoral',
    ]:
        raise ValueError(f'Invalid input: `{administration}` as administration is not supported.')

    pop_cl = Parameter('POP_CL', cl_init, lower=0)
    pop_vc = Parameter('POP_VC', vc_init, lower=0)
    iiv_cl = Parameter('IIV_CL', 0.1)
    iiv_vc = Parameter('IIV_VC', 0.1)

    params = Parameters((pop_cl, pop_vc, iiv_cl, iiv_vc))

    eta_cl_name = 'ETA_CL'
    eta_cl = NormalDistribution.create(eta_cl_name, 'iiv', 0, iiv_cl.symbol)
    eta_vc_name = 'ETA_VC'
    eta_vc = NormalDistribution.create(eta_vc_name, 'iiv', 0, iiv_vc.symbol)
    rvs = RandomVariables.create([eta_cl, eta_vc])

    CL = sympy.Symbol('CL')
    VC = sympy.Symbol('VC')
    cl_ass = Assignment(CL, pop_cl.symbol * sympy.exp(sympy.Symbol(eta_cl_name)))
    vc_ass = Assignment(VC, pop_vc.symbol * sympy.exp(sympy.Symbol(eta_vc_name)))

    cb = CompartmentalSystemBuilder()
    # FIXME: This shouldn't be used here
    from pharmpy.model.external.nonmem.advan import dosing, find_dose

    doses = dosing(di, df, 1)
    central = Compartment.create('CENTRAL', doses=find_dose(doses, 1))
    cb.add_compartment(central)
    cb.add_flow(central, output, CL / VC)

    ipred = Assignment(sympy.Symbol('IPRED'), central.amount / VC)
    y_ass = Assignment(sympy.Symbol('Y'), ipred.symbol)

    stats = Statements([cl_ass, vc_ass, CompartmentalSystem(cb), ipred, y_ass])

    est = EstimationStep.create(
        "FOCE",
        interaction=True,
        maximum_evaluations=99999,
        predictions=['PRED', 'CIPREDI'],
        residuals=['CWRES'],
    )
    eststeps = EstimationSteps.create([est])

    model = Model.create(
        name='start',
        statements=stats,
        estimation_steps=eststeps,
        dependent_variables={y_ass.symbol: 1},
        random_variables=rvs,
        parameters=params,
        description='Start model',
        filename_extension='.mod',  # Should this really be needed?
        dataset=df,
        datainfo=di,
    )

    model = set_proportional_error_model(model)
    model = create_joint_distribution(model, [eta_cl_name, eta_vc_name])

    if administration == 'oral' or administration == 'ivoral':
        model = set_first_order_absorption(model)
        model = set_initial_estimates(model, {'POP_MAT': mat_init})
        model = add_iiv(model, list_of_parameters='MAT', expression='exp', initial_estimate=0.1)
    if administration == 'ivoral':
        # FIXME: Dependent on CMT column having 1 and 2 as values otherwise
        # compartment structure don't match
        model = add_bioavailability(model, logit_transform=True)
        model = set_initial_estimates(model, {'POP_BIO': 0.5})

        # Add IIV to BIO
        # TODO: ? Should there be another initial estimate?
        model = add_iiv(model, list_of_parameters="BIO", expression='add', initial_estimate=0.1)

        # Set dosing to the CENTRAL compartment as well
        ode = model.statements.ode_system
        cb = CompartmentalSystemBuilder(ode)
        if df is None:
            doses = dosing(di, df, 2)
            central_dose = find_dose(doses, comp_number=2, admid=2)
        else:
            # doses = dosing(di, df, 1)
            central_dose = find_dose(doses, comp_number=2, admid=2)
            if not central_dose:
                raise ValueError(
                    (
                        "Could not determine IV dose from dataset. "
                        "Currently require CMT column with values 1 and 2 "
                    )
                )
        cb.set_dose(cb.find_compartment("CENTRAL"), dose=central_dose)

        ode = CompartmentalSystem(cb)
        model = model.replace(
            statements=model.statements.before_odes + ode + model.statements.after_odes
        )
        if df is not None:
            model = add_admid(model)

            # Change bioavailability to piecewise for oral/iv doses
            model = model.replace(
                statements=model.statements.reassign(
                    sympy.Symbol("F_BIO"),
                    sympy.Piecewise(
                        (
                            1 / (1 + sympy.exp(-sympy.Symbol("BIO"))),
                            sympy.Eq(sympy.Symbol('ADMID'), 1),
                        ),
                        (1, sympy.true),
                    ),
                )
            )

        # Add covariate to error model with the following logic
        # RUV = 1 * covariate
        # Y = F + F*EPS*RUV
        ruv_ass = Assignment(sympy.Symbol("RUV"), sympy.Number(1))
        model = model.replace(
            statements=model.statements.before_odes
            + ruv_ass
            + model.statements.ode_system
            + model.statements.after_odes
        )
        Y_ass = model.statements.find_assignment("Y")
        ipred = Y_ass.expression.make_args(Y_ass.expression)[0]
        error = Y_ass.expression.make_args(Y_ass.expression)[1]
        new_Y_ass = Assignment(Y_ass.symbol, ipred + error * ruv_ass.symbol)

        model = model.replace(
            statements=model.statements.reassign(Y_ass.symbol, new_Y_ass.expression)
        )

    return model


def _create_default_datainfo(path_or_df):
    if not isinstance(path_or_df, pd.DataFrame):
        path = path_absolute(path_or_df)
        datainfo_path = path.with_suffix('.datainfo')
        if datainfo_path.is_file():
            di = DataInfo.read_json(datainfo_path)
            di = di.replace(path=path)
            return di
        else:
            with open(path) as file:
                first_line = file.readline()
                if ',' not in first_line:
                    colnames = list(pd.read_csv(path, nrows=0, sep=r'\s+'))
                    separator = r'\s+'
                else:
                    colnames = list(pd.read_csv(path, nrows=0))
                    separator = ','
    else:
        colnames = path_or_df.columns
        separator = None
        path = None

    column_info = []
    for colname in colnames:
        if colname == 'ID' or colname == 'L1':
            info = ColumnInfo.create(colname, type='id', scale='nominal', datatype='int32')
        elif colname == 'DV':
            info = ColumnInfo.create(colname, type='dv')
        elif colname == 'TIME':
            if not set(colnames).isdisjoint({'DATE', 'DAT1', 'DAT2', 'DAT3'}):
                datatype = 'nmtran-time'
            else:
                datatype = 'float64'
            info = ColumnInfo.create(colname, type='idv', scale='ratio', datatype=datatype)
        elif colname == 'EVID':
            info = ColumnInfo.create(colname, type='event', scale='nominal')
        elif colname == 'MDV':
            if 'EVID' in colnames:
                info = ColumnInfo.create(colname, type='mdv')
            else:
                info = ColumnInfo.create(colname, type='event', scale='nominal', datatype='int32')
        elif colname == 'AMT':
            info = ColumnInfo.create(colname, type='dose', scale='ratio')
        elif colname == 'BLQ':
            info = ColumnInfo.create(colname, type='blq', scale='nominal', datatype='int32')
        elif colname == 'LLOQ':
            info = ColumnInfo.create(colname, type='lloq', scale='ratio')
        else:
            info = ColumnInfo.create(colname)
        column_info.append(info)
    di = DataInfo.create(column_info, path=path, separator=separator)
    return di



import pytest

from pharmpy.model import ModelSyntaxError, Parameter
from pharmpy.model.external.nonmem.parsing import parse_thetas
from pharmpy.modeling import set_lower_bounds, set_upper_bounds

INF = float("inf")


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,expected',
    [
        ('$THETA 0 FIX', [(None, 0, -INF, INF, True)]),
        ('$THETA (0,1,INF)', [(None, 1, 0, INF, False)]),
        ('$THETA    12.3 \n\n', [(None, 12.3, -INF, INF, False)]),
        ('$THETA  (0,0.00469) ; CL', [('CL', 0.00469, 0, INF, False)]),
        (
            '$THETA  (0,3) 2 FIXED (0,.6,1) 10 (-INF,-2.7,0)  (37 FIXED)\n 19 (0,1,2)x3',
            [
                (None, 3, 0, INF, False),
                (None, 2, -INF, INF, True),
                (None, 0.6, 0, 1, False),
                (None, 10, -INF, INF, False),
                (None, -2.7, -INF, 0, False),
                (None, 37, -INF, INF, True),
                (None, 19, -INF, INF, False),
                (None, 1, 0, 2, False),
                (None, 1, 0, 2, False),
                (None, 1, 0, 2, False),
            ],
        ),
        (
            '$THETA (0.00469555) FIX ; CL',
            [('CL', 0.00469555, -INF, INF, True)],
        ),
        (
            '$THETA\n ;; Model characteristics\n (0, 0.15, 0.6) ; Proportional\n',
            [("Proportional", 0.15, 0, 0.6, False)],
        ),
        ('$THETA\n;\n1;COMMENTING', [('COMMENTING', 1, -INF, INF, False)]),
        (
            '$THETA\n   ;CMT1;SAMELINE\n;ONLYCOMMENT\n\t;COMMENT2\n    1    \n',
            [(None, 1, -INF, INF, False)],
        ),
        (
            '$THETA\n ;; Model characteristics\n  (0, 0.15, 0.6) ; Proportional error (Drug123)\n'
            '  (0, 1, 10)     ; Additive error (Drug123)\n',
            [('Proportional', 0.15, 0, 0.6, False), ('Additive', 1, 0, 10, False)],
        ),
        ('$THETA  (FIX FIX 0.4) ; CL', [('CL', 0.4, -INF, INF, True)]),
        ('$THETA (FIX 1,   FIX FIX    1 FIX, 1  FIX) ; CL', [('CL', 1, 1, 1, True)]),
        (
            '$THETA\n(0,0.105,)   ; RUV_CVFPG\n',
            [
                ('RUV_CVFPG', 0.105, 0, INF, False),
            ],
        ),
        (
            '$THETA\n(0,0.105,)   ; RUV_CVFPG\n',
            [
                ('RUV_CVFPG', 0.105, 0, INF, False),
            ],
        ),
        (
            '$THETA  (0,3) ; CL\n 2 FIXED ; V\n',
            [
                ('CL', 3, 0, INF, False),
                ('V', 2, -INF, INF, True),
            ],
        ),
        ('$THETA (1,1,1)\n', [(None, 1.0, 1.0, 1.0, True)]),
        ('$THETA (0,1) FIX\n', [(None, 1.0, 0.0, INF, True)]),
        ('$THETA (1,1 FIX)\n', [(None, 1.0, 1.0, INF, True)]),
        ('$THETA (1,1,1 FIX)\n', [(None, 1.0, 1.0, 1.0, True)]),
        ('$THETA 20000000\n', [(None, 20000000, -INF, INF, False)]),
    ],
)
def test_parameters(parser, buf, expected):
    control_stream = parser.parse("$PROB\n" + buf)
    names, bounds, inits, fixs = parse_thetas(control_stream)

    correct_names = [name for name, _, _, _, _ in expected]
    assert names == correct_names

    correct_inits = [init for _, init, _, _, _ in expected]
    assert inits == correct_inits

    correct_lower = [lower for _, _, lower, _, _ in expected]
    lower = [b[0] for b in bounds]
    assert lower == correct_lower

    correct_upper = [upper for _, _, _, upper, _ in expected]
    upper = [b[1] for b in bounds]
    assert upper == correct_upper

    correct_fix = [fix for _, _, _, _, fix in expected]
    assert fixs == correct_fix


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf',
    [
        ('$THETA (-1000001,0)'),
        ('$THETA (0,1,2000000)'),
        ('$THETA (0,1 FIX)'),
        ('$THETA (0,1,1 FIX)'),
        ('$THETA 0'),
        ('$THETA -1000000'),
        ('$THETA (0,1000000)'),
        ('$THETA (1,1)'),
    ],
)
def test_bad_thetas(parser, buf):
    control_stream = parser.parse("$PROB\n" + buf)
    with pytest.raises(ModelSyntaxError):
        _ = parse_thetas(control_stream)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,params,expected',
    [
        ('$THETA 1', [Parameter('THETA_1', 41)], '$THETA 41'),
        ('$THETA 1 FIX', [Parameter('THETA_1', 1, fix=False)], '$THETA 1'),
        (
            '$THETA (2, 2, 2) FIX',
            [Parameter('THETA_1', 2, lower=2, upper=2, fix=False)],
            '$THETA (2, 2, 2)',
        ),
        (
            '$THETA (2, 2, 2 FIX)',
            [Parameter('THETA_1', 2, lower=2, upper=2, fix=False)],
            '$THETA (2, 2, 2)',
        ),
        (
            '$THETA (2, 2, 2)',
            [Parameter('THETA_1', 2, lower=2, upper=2, fix=True)],
            '$THETA (2, 2, 2) FIX',
        ),
        (
            '$THETA 1 2 3 ;CMT',
            [
                Parameter('THETA_1', 1, fix=True),
                Parameter('THETA_2', 2),
                Parameter('THETA_3', 3, fix=True),
            ],
            '$THETA 1 FIX 2 3 FIX ;CMT',
        ),
    ],
)
def test_update(parser, buf, params, expected):
    rec = parser.parse(buf).records[0]
    rec = rec.update(params)
    assert str(rec) == expected


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,remove,expected',
    [
        ('$THETA 1 2 3 ;CMT', [0, 2], '$THETA  2  ;CMT'),
    ],
)
def test_remove_theta(parser, buf, remove, expected):
    rec = parser.parse(buf).records[0]
    rec = rec.remove(remove)
    assert str(rec) == expected


def test_update_lower_bounds(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem/pheno.mod')
    model = set_lower_bounds(model, {'TVCL': -1})
    model = set_upper_bounds(model, {'TVCL': 30})
    expected = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V

$ERROR
Y=F+F*EPS(1)

$THETA (-1,0.00469307,30) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION\n"""
    assert model.model_code == expected

    model = set_lower_bounds(model, {'TVCL': 0})
    expected = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307,30) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION\n"""
    assert model.model_code == expected
